from __future__ import annotations

import pandas as pd
from pandas import DataFrame
from xlwings.constants import Direction

from xlviews.config import rcParams
from xlviews.core.formula import AGG_FUNCS
from xlviews.core.range_collection import RangeCollection
from xlviews.style import set_font, set_number_format
from xlviews.utils import iter_columns, suspend_screen_updates

from .groupby import GroupBy
from .sheet_frame import SheetFrame


class StatsFrame(SheetFrame):
    @suspend_screen_updates
    def __init__(
        self,
        parent: SheetFrame,
        funcs: str | list[str] | None = None,
        by: str | list[str] | None = None,
        *,
        default: str = "median",
        func_column_name: str = "func",
        auto_filter: bool = True,
    ) -> None:
        """Create a StatsFrame.

        Args:
            parent (SheetFrame): The `SheetFrame` to be aggregated.
            funcs (str, list of str, optional): The aggregation
                functions to be used. The following functions are supported:
                    'count', 'sum', 'min', 'max', 'mean', 'median', 'std', 'soa%'
                None to use the default functions.
            by (str, list of str, optional): The column names to be grouped by.
            default (str, optional): The default function to be displayed.
            func_column_name (str, optional): The name of the function column.
            auto_filter (bool, optional): Whether to automatically filter the data.
        """
        funcs = get_func(funcs)
        by = get_by(parent, by)
        offset = get_length(parent, by, funcs) + 2

        # Store the position of the parent SheetFrame before moving down.
        row = parent.row
        column = parent.column
        if isinstance(funcs, list):
            column -= 1

        move_down(parent, offset)

        gp = GroupBy(parent, by)
        data = get_frame(gp, funcs, func_column_name)

        super().__init__(row, column, data, parent.sheet)

        self.as_table(autofit=False, const_header=True)
        self.style()

        if isinstance(funcs, list):
            set_style(self, parent, func_column_name)

        self.alignment("left")

        if self.table and auto_filter and isinstance(funcs, list) and len(funcs) > 1:
            func = default if default in funcs else funcs[0]
            self.table.auto_filter(func_column_name, func)


def get_func(func: str | list[str] | None) -> list[str]:
    if func is None:
        func = list(AGG_FUNCS.keys())
        func.remove("sum")
        func.remove("std")
        return func

    return [func] if isinstance(func, str) else func


def get_by(sf: SheetFrame, by: str | list[str] | None) -> list[str]:
    if not by:
        return sf.index.names

    return list(iter_columns(sf.index.names, by))


def get_length(sf: SheetFrame, by: list[str], funcs: list | dict) -> int:
    n = 1 if isinstance(funcs, dict) else len(funcs)

    if not by:
        return n

    return len(sf.index.to_frame()[by].drop_duplicates()) * n


def get_frame(
    group: GroupBy,
    funcs: list[str],
    func_column_name: str = "func",
) -> DataFrame:
    df = group.agg(funcs, group.sf.columns.to_list(), formula=True, as_address=True)
    df = df.stack(level=1, future_stack=True)  # noqa: PD013

    index = df.index.to_frame()
    index = index.rename(columns={index.columns[-1]: func_column_name})

    for c in group.sf.index.names:
        if c not in group.by:
            index[c] = ""

    df = pd.concat([index, df], axis=1)
    return df.set_index([func_column_name, *group.sf.index.names])


def has_header(sf: SheetFrame) -> bool:
    start = sf.cell.offset(-1)
    end = start.offset(0, sf.index.nlevels)
    value = sf.sheet.range(start, end).options(ndim=1).value

    if not isinstance(value, list):
        raise NotImplementedError

    return any(value)


def move_down(sf: SheetFrame, length: int) -> int:
    start = sf.row - 1
    end = sf.row + length - 2

    if has_header(sf):
        end += 1

    rows = sf.sheet.api.Rows(f"{start}:{end}")
    rows.Insert(Shift=Direction.xlDown)
    sf.cell = sf.cell.offset()  # update cell
    return end - start + 1


def set_style(sf: SheetFrame, parent: SheetFrame, func_column_name: str) -> None:
    idx = [sf.column + i for i in range(sf.index.nlevels + len(sf.columns))]

    columns = (*parent.index.names, *parent.columns)
    formats = [None, *[parent.get_number_format(column) for column in columns]]

    for (func,), rows in sf.groupby(func_column_name).items():
        for col, fmt in zip(idx, formats, strict=True):
            rc = RangeCollection(rows, col, sf.sheet)

            if func in ["median", "min", "mean", "max", "std", "sum"] and fmt:
                set_number_format(rc, fmt)

            color = rcParams.get(f"stats.{func}.color")
            italic = rcParams.get(f"stats.{func}.italic")
            set_font(rc, color=color, italic=italic)

            if func == "soa" and col != idx[0]:
                set_number_format(rc, "0.0%")

    rng = sf.get_range(func_column_name)
    set_font(rng, italic=True)



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.testing.common import create_sheet
from xlviews.testing.heat_frame.common import HeatFrameContainer
from xlviews.testing.sheet_frame.pivot import Base as BaseParent
from xlviews.testing.sheet_frame.pivot import MultiIndex as MultiIndexParent_

if TYPE_CHECKING:
    from pandas import DataFrame

    from xlviews.dataframes.sheet_frame import SheetFrame


class Base(HeatFrameContainer):
    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        return sf.pivot_table("v", "y", "x", formula=True)

    def init(self) -> None:
        cb = self.sf.colorbar(label="v", autofit=True)
        cb.set_adjacent_column_width(1)
        self.sf.style(color="#e0e0e0").autofit()


class MultiIndexParent(MultiIndexParent_):
    column: int = 15

    def init(self) -> None:
        self.sf.number_format("0.0").autofit()


class MultiIndex(HeatFrameContainer):
    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        return sf.pivot_table("v", ["Y", "y"], ["X", "x"], formula=True)


if __name__ == "__main__":
    sheet = create_sheet()

    fc = BaseParent(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)
    fc = Base(sf)
    fc.sf.set_adjacent_column_width(1)

    fc = MultiIndexParent(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)
    fc = MultiIndex(sf)
    fc.sf.set_adjacent_column_width(1)



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING, Any

import xlwings
from pywintypes import com_error
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.style import hide_gridlines

if TYPE_CHECKING:
    from pandas import DataFrame
    from xlwings import Sheet


@cache
def is_app_available() -> bool:
    try:
        with xlwings.App(visible=False):
            pass
    except com_error:
        return False

    return True


def create_sheet() -> Sheet:
    for app in xlwings.apps:
        app.quit()

    book = xlwings.Book()
    sheet = book.sheets.add()
    sheet.range("A1").column_width = 1
    hide_gridlines(sheet)

    return sheet


def create_sheet_frame(
    df: DataFrame,
    sheet: Sheet,
    row: int,
    column: int,
    style: bool = False,
    table: bool = False,
) -> SheetFrame:
    sf = SheetFrame(row, column, df, sheet)
    if style:
        sf.style()
        sf.autofit()

    if table:
        sf.as_table()

    return sf


class FrameContainer:
    df: DataFrame
    sf: SheetFrame
    row: int = 2
    column: int = 2

    def __init__(
        self,
        sheet: Sheet,
        row: int = 0,
        column: int = 0,
        style: bool = False,
        **kwargs,
    ) -> None:
        self.df = self.dataframe()
        self.row = row or self.row
        self.column = column or self.column
        self.sf = create_sheet_frame(
            self.df,
            sheet,
            self.row,
            self.column,
            style=style,
            **self.kwargs(**kwargs),
        )
        self.init()

    def kwargs(self, **kwargs) -> dict[str, Any]:
        return kwargs

    def init(self) -> None:
        pass

    @classmethod
    def dataframe(cls) -> DataFrame:
        raise NotImplementedError



import pytest


@pytest.mark.parametrize(
    ("name", "color"),
    [
        ("black", 0),
        ("red", 2**8 - 1),
        ("green", 32768),
        ("blue", 256 * 256 * 255),
        ("white", 2**24 - 1),
        ("aqua", 16776960),
        ("gray", 8421504),
        ("indigo", 8519755),
        ("lime", 65280),
        ("pink", 13353215),
        ((10, 10, 10), 10 + 10 * 256 + 10 * 256 * 256),
        (100, 100),
    ],
)
def test_rgb(name, color):
    from xlviews.colors import rgb

    assert rgb(name) == color
    if isinstance(name, tuple):
        assert rgb(*name) == color


@pytest.mark.parametrize("name", ["invalid", (1, "x", "y")])
def test_rgb_error(name):
    from xlviews.colors import rgb

    with pytest.raises(ValueError, match="Invalid color format"):
        rgb(name)



import pytest
from pandas import DataFrame

from xlviews.dataframes.groupby import GroupBy
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def gr(sf_parent: SheetFrame):
    return GroupBy(sf_parent, ["x", "y"])


@pytest.mark.parametrize(
    ("funcs", "shape"),
    [(["mean"], (4, 3)), (["min", "max"], (8, 3))],
)
def test_get_frame_shape(gr: GroupBy, funcs, shape):
    from xlviews.dataframes.stats_frame import get_frame

    assert get_frame(gr, funcs).shape == shape


def test_get_frame_index_list(gr: GroupBy):
    from xlviews.dataframes.stats_frame import get_frame

    df = get_frame(gr, ["mean"])
    assert df.index.names == ["func", "x", "y", "z"]


def test_get_frame_offset(gr: GroupBy):
    from xlviews.dataframes.stats_frame import get_frame

    df = get_frame(gr, ["mean"]).reset_index()
    assert df["x"].iloc[0] == "=$C$4"
    assert df["y"].iloc[-1] == "=$D$16"
    assert df["a"].iloc[0] == "=AGGREGATE(1,7,$F$4:$F$7,$F$20:$F$23)"
    assert df["c"].iloc[-1] == "=AGGREGATE(1,7,$H$16:$H$19)"


def test_get_by_none(sf_parent: SheetFrame):
    from xlviews.dataframes.stats_frame import get_by

    assert get_by(sf_parent, None) == ["x", "y", "z"]



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.sheet_frame.pivot import Base

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc_parent(sheet_module: Sheet):
    return Base(sheet_module)


@pytest.fixture(scope="module")
def df_parent(fc_parent: Base):
    return fc_parent.df


@pytest.fixture(scope="module")
def sf_parent(fc_parent: Base):
    return fc_parent.sf


@pytest.fixture(params=["u", "v", None, ["v", "u"]])
def values(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(params=["y", ["y"]])
def index(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(params=["x", ["x"]])
def columns(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def sf(sf_parent: SheetFrame, values, index, columns, sheet):
    df = sf_parent.pivot_table(
        values,
        index,
        columns,
        include_sheetname=True,
        formula=True,
    )
    return SheetFrame(2, 2, df, sheet=sheet)


@pytest.fixture
def df(df_parent: DataFrame, values, index, columns):
    return df_parent.pivot_table(values, index, columns, aggfunc=lambda x: x)


def test_index(sf: SheetFrame, df: DataFrame):
    assert sf.value.index.equals(df.index)


def test_columns(sf: SheetFrame, df: DataFrame):
    assert sf.value.columns.equals(df.columns)


def test_values(sf: SheetFrame, df: DataFrame):
    assert sf.value.equals(df)


@pytest.fixture(params=[("y", "x"), ("y", None), (None, "x")])
def index_columns(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(params=["mean", ["sum", "count"]])
def aggfunc(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def sf_agg(sf_parent: SheetFrame, values, index_columns, aggfunc, sheet):
    index, columns = index_columns
    df = sf_parent.pivot_table(
        values,
        index,
        columns,
        aggfunc,
        include_sheetname=True,
        formula=True,
    )
    return SheetFrame(2, 2, df, sheet=sheet)


@pytest.fixture
def df_agg(df_parent: DataFrame, values, index_columns, aggfunc):
    index, columns = index_columns
    return df_parent.pivot_table(values, index, columns, aggfunc)


def test_agg_shape(sf_agg: SheetFrame, df_agg: DataFrame):
    assert sf_agg.value.shape == df_agg.shape


def test_agg_index(sf_agg: SheetFrame, df_agg: DataFrame):
    assert sf_agg.value.index.equals(df_agg.index)


def test_agg_columns(sf_agg: SheetFrame, df_agg: DataFrame):
    assert sf_agg.value.columns.equals(df_agg.columns)


def test_agg_values(sf_agg: SheetFrame, df_agg: DataFrame):
    assert sf_agg.value.equals(df_agg.astype(float))


@pytest.mark.parametrize("aggfunc", [None, "mean"])
def test_error(sf_parent: SheetFrame, aggfunc):
    with pytest.raises(ValueError, match="No group keys passed!"):
        sf_parent.pivot_table(None, None, None, aggfunc, formula=True)



import numpy as np
import pytest
from xlwings import Range as RangeImpl
from xlwings import Sheet

from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module", params=["C5", "D10:D13", "F4:I4", "C5:E9"])
def addr(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(scope="module")
def rng_impl(addr, sheet_module: Sheet):
    return sheet_module.range(addr)


@pytest.fixture(scope="module", params=[True, False])
def include_sheetname(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(scope="module", params=[True, False])
def external(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(scope="module")
def addr_impl(rng_impl: RangeImpl, include_sheetname, external):
    return rng_impl.get_address(include_sheetname=include_sheetname, external=external)


def test_range_int_int(rng_impl: RangeImpl, include_sheetname, external):
    rng = Range(rng_impl.row, rng_impl.column)
    x = rng.get_address(include_sheetname=include_sheetname, external=external)
    y = rng_impl[0].get_address(include_sheetname=include_sheetname, external=external)
    assert x == y


def test_range_tuple_first(rng_impl: RangeImpl, include_sheetname, external):
    rng = Range((rng_impl.row, rng_impl.column))
    x = rng.get_address(include_sheetname=include_sheetname, external=external)
    y = rng_impl[0].get_address(include_sheetname=include_sheetname, external=external)
    assert x == y


def test_range_tuple_last(rng_impl: RangeImpl, include_sheetname, external):
    rng = Range((rng_impl.last_cell.row, rng_impl.last_cell.column))
    x = rng.get_address(include_sheetname=include_sheetname, external=external)
    y = rng_impl[-1].get_address(include_sheetname=include_sheetname, external=external)
    assert x == y


def test_range_tuple_tuple(rng_impl: RangeImpl, addr_impl, include_sheetname, external):
    cell1 = (rng_impl.row, rng_impl.column)
    cell2 = (rng_impl.last_cell.row, rng_impl.last_cell.column)
    rng = Range(cell1, cell2)
    x = rng.get_address(include_sheetname=include_sheetname, external=external)
    assert x == addr_impl


def test_range_from_range(rng_impl: RangeImpl, addr_impl, include_sheetname, external):
    rng = Range.from_range(rng_impl)
    x = rng.get_address(include_sheetname=include_sheetname, external=external)
    assert x == addr_impl


def test_range_from_range_first(rng_impl: RangeImpl):
    rng = Range.from_range(rng_impl[0])
    assert rng.get_address() == rng_impl[0].get_address()


def test_range_from_range_last(rng_impl: RangeImpl):
    rng = Range.from_range(rng_impl.last_cell)
    assert rng.get_address() == rng_impl.last_cell.get_address()


def test_range_error_tuple(sheet_module: Sheet):
    with pytest.raises(TypeError, match="cell2 must be a tuple or None"):
        Range((1, 1), 1)


def test_range_error_int(sheet_module: Sheet):
    with pytest.raises(TypeError, match="cell2 must be an integer"):
        Range(1, (1, 1))


@pytest.fixture(scope="module")
def rng(rng_impl: RangeImpl):
    return Range.from_range(rng_impl)


def test_len(rng: Range, rng_impl: RangeImpl):
    assert len(rng) == len(rng_impl)


def test_iter(rng: Range, rng_impl: RangeImpl):
    x = [r.get_address() for r in rng]
    y = [r.get_address() for r in rng_impl]
    assert x == y


def test_getitem(rng: Range, rng_impl: RangeImpl):
    for k in range(len(rng)):
        assert rng[k].get_address() == rng_impl[k].get_address()


def test_getitem_neg(rng: Range, rng_impl: RangeImpl):
    for k in range(len(rng)):
        assert rng[-k].get_address() == rng_impl[-k].get_address()


def test_getitem_error(rng: Range):
    with pytest.raises(IndexError, match="Index out of range"):
        rng[100]


def test_repr(rng: Range, rng_impl: RangeImpl):
    assert repr(rng) == repr(rng_impl)


def test_last_cell(rng: Range, rng_impl: RangeImpl):
    assert rng.last_cell.get_address() == rng_impl.last_cell.get_address()


@pytest.mark.parametrize("row_offset", [2, 0, -2])
@pytest.mark.parametrize("column_offset", [2, 0, -2])
def test_offset(rng: Range, rng_impl: RangeImpl, row_offset, column_offset):
    x = rng.offset(row_offset, column_offset)
    y = rng_impl.offset(row_offset, column_offset)
    assert x.get_address() == y.get_address()


def test_impl_from(rng: Range, rng_impl: RangeImpl):
    rng_impl.value = rng_impl.get_address(external=True)
    assert rng_impl.value == rng.impl.value


def test_impl_to(rng: Range, rng_impl: RangeImpl):
    rng.impl.value = rng.get_address()
    assert rng_impl.value == rng.impl.value


def test_iter_addresses(rng: Range, rng_impl: RangeImpl, external):
    x = list(rng.iter_addresses(external=external))
    y = [r.get_address(external=external) for r in rng_impl]
    assert x == y


def test_iter_addresses_formula(rng: Range, rng_impl: RangeImpl, external):
    x = list(rng.iter_addresses(external=external, formula=True))
    y = ["=" + r.get_address(external=external) for r in rng_impl]
    assert x == y


def test_frame_range_cell(sheet_module: Sheet):
    rng = Range((1, 1), (1, 1), sheet_module).frame
    df = rng.get_address()
    assert df.shape == (1, 1)
    assert df.index.tolist() == [0]
    assert df.columns.tolist() == [0]
    np.testing.assert_array_equal(df, [["$A$1"]])


def test_frame_range_row(sheet_module: Sheet):
    rng = Range((1, 1), (1, 3), sheet_module).frame
    df = rng.get_address(row_absolute=False)
    assert df.shape == (1, 3)
    assert df.index.tolist() == [0]
    assert df.columns.tolist() == [0, 1, 2]
    np.testing.assert_array_equal(df, [["$A1", "$B1", "$C1"]])


def test_frame_range_column(sheet_module: Sheet):
    rng = Range((1, 1), (3, 1), sheet_module).frame
    df = rng.get_address(column_absolute=False)
    assert df.shape == (3, 1)
    assert df.index.tolist() == [0, 1, 2]
    assert df.columns.tolist() == [0]
    np.testing.assert_array_equal(df, [["A$1"], ["A$2"], ["A$3"]])


def test_frame_range_matrix(sheet_module: Sheet):
    rng = Range((1, 1), (3, 2), sheet_module).frame
    df = rng.get_address(column_absolute=False, row_absolute=False, formula=True)
    assert df.shape == (3, 2)
    assert df.index.tolist() == [0, 1, 2]
    assert df.columns.tolist() == [0, 1]
    np.testing.assert_array_equal(df, [["=A1", "=B1"], ["=A2", "=B2"], ["=A3", "=B3"]])


def test_frame_range_cell_sheetname(sheet_module: Sheet):
    rng = Range((1, 1), (1, 1), sheet_module).frame
    df = rng.get_address(include_sheetname=True)
    assert df.loc[0, 0] == f"{sheet_module.name}!$A$1"


def test_frame_range_cell_external(sheet_module: Sheet):
    rng = Range((1, 1), (1, 1), sheet_module).frame
    df = rng.get_address(external=True)
    assert df.loc[0, 0] == f"[{sheet_module.book.name}]{sheet_module.name}!$A$1"



import pytest
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.dataframes.table import Table
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import Index

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Index(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


@pytest.fixture
def table(sf: SheetFrame):
    yield sf.as_table()
    sf.unlist()


@pytest.mark.parametrize("value", ["x", "y"])
def test_table(table: Table, value):
    table.auto_filter("name", value)
    header = table.const_header.value
    assert isinstance(header, list)
    assert header[0] == value



import pytest
from xlwings import Sheet

from xlviews.testing import FrameContainer
from xlviews.testing.stats_frame import Parent


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Parent(sheet_module, 3, 3, table=True)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf_parent(fc: FrameContainer):
    return fc.sf



def test_marker_style_int():
    from xlviews.chart.style import get_marker_style

    assert get_marker_style(1) == 1


def test_line_style_int():
    from xlviews.chart.style import get_line_style

    assert get_line_style(1) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.chart.axes import Axes
from xlviews.figure.palette import get_color_palette, get_marker_palette
from xlviews.figure.plot import Plot
from xlviews.testing.common import create_sheet
from xlviews.testing.sheet_frame.pivot import Pivot

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

    from xlviews.dataframes.sheet_frame import SheetFrame


def facet(sf: SheetFrame) -> Iterator[tuple[dict[str, Any], Plot]]:
    sf.set_adjacent_column_width(1)

    ax = Axes(2, 11)
    data = sf.groupby(["A", "B", "X"]).agg(include_sheetname=True)
    cp = get_color_palette(data, "X")
    mp = get_marker_palette(data, "A")

    for key, plot in Plot.facet(ax, data, index="B", columns="A"):
        plot.add("u", "v").set(color=cp, marker=mp, label="{X}", alpha=0.8)
        plot.axes.title = "{A}_{B}".format(**key)
        yield key, plot


if __name__ == "__main__":
    sheet = create_sheet()
    fc = Pivot(sheet, style=True)
    list(facet(fc.sf))



from itertools import islice

import pytest
from pandas import DataFrame


@pytest.fixture(scope="module")
def df():
    return DataFrame(
        {
            "a": [1, 1, 2, 2, 3, 3],
            "b": [4, 4, 5, 6, 7, 7],
            "c": [10, 10, 10, 11, 11, 11],
        },
    )


def test_get_columns_default(df: DataFrame):
    from xlviews.figure.palette import get_columns_default

    columns, default = get_columns_default(df, "a")
    assert columns == ["a"]
    assert default == {}


def test_get_columns_default_with_default_list(df: DataFrame):
    from xlviews.figure.palette import get_columns_default

    columns, default = get_columns_default(df, "a", ["red", "blue"])
    assert columns == ["a"]
    assert default[(1,)] == "red"
    assert default[(2,)] == "blue"
    assert default[(3,)] == "red"


@pytest.mark.parametrize(
    ("columns", "index"),
    [
        (["a"], {(1,): 0, (2,): 1, (3,): 2}),
        (["b"], {(4,): 0, (5,): 1, (6,): 2, (7,): 3}),
        (["c"], {(10,): 0, (11,): 1}),
        (["a", "b"], {(1, 4): 0, (2, 5): 1, (2, 6): 2, (3, 7): 3}),
        (["a", "c"], {(1, 10): 0, (2, 10): 1, (2, 11): 2, (3, 11): 3}),
        (["b", "c"], {(4, 10): 0, (5, 10): 1, (6, 11): 2, (7, 11): 3}),
    ],
)
def test_get_index(df: DataFrame, columns, index):
    from xlviews.figure.palette import get_index

    assert get_index(df[columns]) == index


@pytest.mark.parametrize(
    ("columns", "default", "index"),
    [
        (["a"], [3, 2], {(1,): 2, (2,): 1, (3,): 0}),
        (["b"], [6, 7], {(4,): 2, (5,): 3, (6,): 0, (7,): 1}),
        (["c"], [12, 11], {(10,): 1, (11,): 0}),
        (["a", "b"], [[2, 5], [3, 7]], {(1, 4): 2, (2, 5): 0, (2, 6): 3, (3, 7): 1}),
        (
            ["a", "c"],
            [[2, 10], (3, 11)],
            {(1, 10): 2, (2, 10): 0, (2, 11): 3, (3, 11): 1},
        ),
        (["b", "c"], [], {(4, 10): 0, (5, 10): 1, (6, 11): 2, (7, 11): 3}),
    ],
)
def test_get_index_default(df: DataFrame, columns, default, index):
    from xlviews.figure.palette import get_index

    assert get_index(df[columns], default) == index


def test_cycle_colors():
    from xlviews.figure.palette import cycle_colors

    x = list(islice(cycle_colors(), 3))
    assert x == ["#1f77b4", "#ff7f0e", "#2ca02c"]


def test_cycle_colors_skips():
    from xlviews.figure.palette import cycle_colors

    x = list(islice(cycle_colors(["#ff7f0e"]), 3))
    assert x == ["#1f77b4", "#2ca02c", "#d62728"]


def test_cycle_markers():
    from xlviews.figure.palette import cycle_markers

    x = list(islice(cycle_markers(), 3))
    assert x == ["o", "^", "s"]


def test_cycle_markers_skips():
    from xlviews.figure.palette import cycle_markers

    x = list(islice(cycle_markers(["o"]), 3))
    assert x == ["^", "s", "d"]


@pytest.mark.parametrize(
    ("key", "value"),
    [(1, 0), ((1,), 0), (3, 2), ((3,), 2)],
)
def test_marker_palette_get(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, "a")
    assert p.get(key) == value


@pytest.mark.parametrize(
    ("key", "value"),
    [({"a": 1}, "o"), ({"a": 2}, "^"), ({"a": 3}, "s")],
)
def test_marker_palette(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, "a")
    assert p[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [({"a": 1, "b": 0}, "o"), ({"x": 0, "a": 2}, "^"), ({"a": 3}, "s")],
)
def test_marker_palette_dict(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, "a")
    assert p[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [({"a": 1, "b": 0}, "^"), ({"a": 2, "b": 0}, "o"), ({"a": 3, "b": 0}, "x")],
)
def test_marker_palette_default(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, "a", {2: "o", 3: "x"})
    assert p[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [({"a": 1, "b": 4}, "o"), ({"a": 2, "b": 5}, "^"), ({"a": 2, "b": 6}, "s")],
)
def test_marker_palette_multi(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, ["a", "b"])
    assert p[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [({"a": 1, "b": 4}, "o"), ({"a": 2, "b": 5}, "x"), ({"a": 2, "b": 6}, "o")],
)
def test_marker_palette_multi_default_list(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, ["a", "b"], ["o", "x"])
    assert p[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [
        ({"a": 1, "b": 4, "c": 2}, "o"),
        ({"a": 2, "b": 5, "c": 2}, "^"),
        ({"a": 2, "b": 6, "c": 2}, "s"),
    ],
)
def test_marker_palette_multi_dict(df: DataFrame, key, value):
    from xlviews.figure.palette import MarkerPalette

    p = MarkerPalette(df, ["a", "b"])
    assert p[key] == value


@pytest.mark.parametrize(("key", "value"), [(4, "#1f77b4"), (5, "red"), (6, "blue")])
def test_color_palette(df: DataFrame, key, value):
    from xlviews.figure.palette import ColorPalette

    p = ColorPalette(df, "b", {5: "red", 6: "blue", 7: "green"})
    assert p[{"b": key}] == value


@pytest.mark.parametrize(("key", "value"), [(4, "red"), (5, "blue"), (6, "red")])
def test_color_palette_default_list(df: DataFrame, key, value):
    from xlviews.figure.palette import ColorPalette

    p = ColorPalette(df, "b", ["red", "blue"])
    assert p[{"b": key}] == value


def test_get_palette_none(df: DataFrame):
    from xlviews.figure.palette import MarkerPalette, get_palette

    assert get_palette(MarkerPalette, df, None) is None


def test_get_palette(df: DataFrame):
    from xlviews.figure.palette import MarkerPalette, get_palette

    p = get_palette(MarkerPalette, df, "a")
    assert isinstance(p, MarkerPalette)
    assert p[{"a": 1}] == "o"
    assert p[{"a": 2}] == "^"
    assert p[{"a": 3}] == "s"

    assert get_palette(MarkerPalette, df, p) is p


def test_get_palette_dict(df: DataFrame):
    from xlviews.figure.palette import MarkerPalette, get_palette

    p = get_palette(MarkerPalette, df, {(1, 4, 10): "x"})
    assert isinstance(p, MarkerPalette)
    assert p[{"a": 1, "b": 4, "c": 10}] == "x"
    assert p[{"a": 2, "b": 5, "c": 10}] == "o"


def test_get_palette_tuple_list(df: DataFrame):
    from xlviews.figure.palette import MarkerPalette, get_palette

    p = get_palette(MarkerPalette, df, ("b", ["o", "x"]))
    assert isinstance(p, MarkerPalette)
    assert p[{"b": 4}] == "o"
    assert p[{"b": 5}] == "x"
    assert p[{"b": 6}] == "o"
    assert p[{"b": 7}] == "x"


def test_get_palette_tuple_dict(df: DataFrame):
    from xlviews.figure.palette import MarkerPalette, get_palette

    p = get_palette(MarkerPalette, df, ("b", {"4": "X", "5": "Y"}))
    assert isinstance(p, MarkerPalette)
    assert p[{"b": 4}] == "X"
    assert p[{"b": 5}] == "Y"
    assert p[{"b": 6}] == "o"
    assert p[{"b": 7}] == "^"


def test_series():
    from pandas import Series

    from xlviews.figure.palette import MarkerPalette, get_palette

    s = Series([1, 2, 3], index=["a", "b", "c"])
    data = s.to_frame().T
    p = get_palette(MarkerPalette, data, "x")
    assert p
    assert p[{None: 0}] == "x"


def test_get_palette_new_default(df: DataFrame):
    from xlviews.figure.palette import ColorPalette, get_palette

    p = get_palette(ColorPalette, df, "red")
    assert p
    assert p[{"a": 1, "b": 4, "c": 10}] == "red"
    assert p[{"a": 2, "b": 5, "c": 10}] == "red"


def test_get_palette_new_default_list(df: DataFrame):
    from xlviews.figure.palette import ColorPalette, get_palette

    p = get_palette(ColorPalette, df, ["red", "blue", "green"])
    assert p
    assert p[{"a": 1, "b": 4, "c": 10}] == "red"
    assert p[{"a": 2, "b": 5, "c": 10}] == "blue"
    assert p[{"a": 2, "b": 6, "c": 11}] == "green"
    assert p[{"a": 3, "b": 7, "c": 11}] == "red"


def test_function_palette_str():
    from xlviews.figure.palette import FunctionPalette

    p = FunctionPalette("a", lambda x: x)
    assert p[{"a": 1}] == 1


def test_function_palette_list():
    from xlviews.figure.palette import FunctionPalette

    p = FunctionPalette(["a"], lambda x: x[0])  # type: ignore
    assert p[{"a": 1}] == 1


def test_get_palette_callable_str(df: DataFrame):
    from xlviews.figure.palette import ColorPalette, get_palette

    df = df.set_index("a")
    p = get_palette(ColorPalette, df, lambda x: str(x))
    assert p
    assert p[{"a": 1}] == "1"


def test_get_palette_callable_list(df: DataFrame):
    from xlviews.figure.palette import ColorPalette, get_palette

    df = df.set_index(["a", "b"])
    p = get_palette(ColorPalette, df, lambda x: str(x))
    assert p
    assert p[{"a": 1, "b": 4}] == "(1, 4)"


def test_get_palette_callable_multi_index(df: DataFrame):
    from xlviews.figure.palette import ColorPalette, get_palette

    df = df.set_index(["a", "b"])
    p = get_palette(ColorPalette, df, ("a", lambda x: str(x)))
    assert p
    assert p[{"a": 1, "b": 4}] == "1"


def test_get_marker_palette_callable(df: DataFrame):
    from xlviews.figure.palette import get_marker_palette

    df = df.set_index(["a", "b"])
    p = get_marker_palette(df, "a")
    assert p
    assert p[{"a": 1, "b": 4}] == "o"
    assert p[{"a": 2, "b": 5}] == "^"
    assert p[{"a": 3, "b": 6}] == "s"


def test_get_color_palette_callable(df: DataFrame):
    from xlviews.figure.palette import get_color_palette

    df = df.set_index(["a", "b"])
    p = get_color_palette(df, "b")
    assert p
    assert p[{"a": 1, "b": 4}] == "#1f77b4"
    assert p[{"a": 2, "b": 5}] == "#ff7f0e"
    assert p[{"a": 3, "b": 6}] == "#2ca02c"



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.core.range import Range
from xlviews.testing.common import create_sheet
from xlviews.testing.heat_frame.common import HeatFrameContainer
from xlviews.testing.sheet_frame.pivot import MultiIndex
from xlviews.utils import add_validate_list

if TYPE_CHECKING:
    from pandas import DataFrame

    from xlviews.dataframes.sheet_frame import SheetFrame


class AggParent(MultiIndex):
    column: int = 2


class AggStr(HeatFrameContainer):
    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        return sf.pivot_table("v", "Y", "X", "mean", formula=True)


class AggRange(HeatFrameContainer):
    row: int = 8

    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        func = Range((13, 14), sheet=sf.sheet)
        add_validate_list(func, ["min", "max", "mean", "median", "soa"], "mean")
        return sf.pivot_table("v", "Y", "X", func, formula=True)


if __name__ == "__main__":
    sheet = create_sheet()

    fc = AggParent(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)
    fc = AggStr(sf)
    fc = AggRange(sf)
    fc.sf.set_adjacent_column_width(1)



from __future__ import annotations

from xlviews.testing.common import create_sheet
from xlviews.testing.sheet_frame.base import Index, MultiColumn, NoIndex

if __name__ == "__main__":
    sheet = create_sheet()
    fc = NoIndex(sheet, style=True)
    print(fc.sf.melt())

    print(fc.sf.agg())

    fc = Index(sheet, style=True)
    print(fc.sf.groupby("name").agg())

    fc = MultiColumn(sheet, style=True)
    print(fc.sf.melt())

    print(fc.sf.agg())



from __future__ import annotations

import numpy as np
from pandas import DataFrame

from xlviews.dataframes.stats_frame import StatsFrame
from xlviews.testing.common import FrameContainer, create_sheet


class Parent(FrameContainer):
    row: int = 3
    column: int = 3

    @classmethod
    def dataframe(cls) -> DataFrame:
        df = DataFrame(
            {
                "x": ["a"] * 8 + ["b"] * 8 + ["a"] * 4,
                "y": (["c"] * 4 + ["d"] * 4) * 2 + ["c"] * 4,
                "z": range(1, 21),
                "a": range(20),
                "b": list(range(10)) + list(range(0, 30, 3)),
                "c": list(range(20, 40, 2)) + list(range(0, 20, 2)),
            },
        )
        df = df.set_index(["x", "y", "z"])
        df.iloc[[4, -1], 0] = np.nan
        df.iloc[[3, 6, 9], -1] = np.nan
        return df

    def init(self) -> None:
        self.sf.number_format(c="0.00")


if __name__ == "__main__":
    sheet = create_sheet()
    fc = Parent(sheet, 3, 3, style=True, table=True)
    sf_parent = fc.sf

    funcs = ["count", "mean", "median", "min", "max", "soa", "sum"]
    StatsFrame(sf_parent, funcs, by=":y")



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.sheet_frame.pivot import Pivot
from xlviews.utils import iter_group_locs

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Pivot(sheet_module)


@pytest.fixture(scope="module")
def sf(fc: Pivot):
    return fc.sf


@pytest.fixture(scope="module")
def df(sf: SheetFrame):
    return sf.pivot_table("u", ["B", "Y"], ["A", "X"], "mean")


def test_index_level_values(df: DataFrame):
    v = df.index.get_level_values(0).to_list()
    assert v == [1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3]


def test_columns_level_values(df: DataFrame):
    v = df.columns.get_level_values(0).to_list()
    assert v == [1, 1, 1, 2, 2]


def test_index_group_locs(df: DataFrame):
    v = list(iter_group_locs(df.index.get_level_values(0)))
    assert v == [(0, 3), (4, 6), (7, 10)]


def test_columns_group_locs(df: DataFrame):
    v = list(iter_group_locs(df.columns.get_level_values(0)))
    assert v == [(0, 2), (3, 4)]



from __future__ import annotations

from itertools import product

import numpy as np
from pandas import DataFrame, Index

from xlviews.core.range import Range
from xlviews.utils import iter_columns, suspend_screen_updates

from .sheet_frame import SheetFrame


class DistFrame(SheetFrame):
    dist_func: dict[str, str]

    @suspend_screen_updates
    def __init__(
        self,
        parent: SheetFrame,
        columns: str | list[str] | None = None,
        by: str | list[str] | None = None,
        dist: str | dict[str, str] = "norm",
    ) -> None:
        if columns is None:
            columns = parent.columns.to_list()
        elif isinstance(columns, str):
            columns = [columns]

        self.dist_func = get_dist_func(dist, columns)

        row = parent.row
        column = parent.column + parent.width + 1
        by = list(iter_columns(parent.index.names, by)) if by else []
        index = select_index(parent.index, by)
        data = get_init_data(index, columns)

        super().__init__(row, column, data, parent.sheet)

        self.set_values(parent, columns, by)
        self.style()
        self.autofit()

    def set_values(
        self,
        parent: SheetFrame,
        columns: list[str],
        by: list[str],
    ) -> None:
        group = parent.groupby(by)
        parent_columns = parent.get_indexer(columns)
        self_columns = self.get_indexer([c + "_n" for c in columns])

        it = zip(parent_columns, self_columns, columns, strict=True)
        for parent_column, self_column, column in it:
            dist = self.dist_func[column]

            for row in group.values():
                if len(row) != 1:
                    raise ValueError("group must be continuous")

                start = row[0][0]
                length = row[0][1] - start + 1

                parent_cell = Range((start, parent_column), sheet=parent.sheet)
                self_cell = Range((start, self_column), sheet=self.sheet)

                formula = counter(parent_cell)
                set_formula(self_cell, length, formula)

                formula = sorted_value(parent_cell, self_cell, length)
                set_formula(self_cell.offset(0, 1), length, formula)

                formula = sigma_value(self_cell, length, dist)
                set_formula(self_cell.offset(0, 2), length, formula)

        idx = parent.get_indexer(columns)
        for column, i in zip(columns, idx, strict=True):
            fmt = parent.sheet.range(parent.row + 1, i).number_format
            self.number_format({f"{column}_v": fmt, f"{column}_s": "0.00"})

    # def plot(
    #     self,
    #     x,
    #     label="auto",
    #     color=None,
    #     marker=None,
    #     axes=None,
    #     xlabel="auto",
    #     ylabel="auto",
    #     **kwargs,
    # ):
    #     if ylabel == "auto":
    #         dist = self.dist_func[x] if isinstance(x, str) else self.dist_func[x[0]]
    #         ylabel = "σ" if dist == "norm" else "ln(-ln(1-F))"
    #     plot = None
    #     if isinstance(x, str) and xlabel == "auto":
    #         x_ = x.split("_")[0]
    #         xlabel = rcParams.get(f"axis.label.{x_}", x)
    #         if "_" in x and "[" in xlabel:
    #             xlabel = x + " " + xlabel[xlabel.index("[") :]
    #     xs = [x] if isinstance(x, str) else x
    #     colors = color if isinstance(color, list) else [color] * len(xs)
    #     markers = marker if isinstance(marker, list) else [marker] * len(xs)
    #     for x_, color, marker in zip(xs, colors, markers, strict=False):
    #         label_ = x_ if label == "auto" and isinstance(x, list) else label
    #         plot = self._plot(
    #             x_,
    #             label=label_,
    #             axes=axes,
    #             color=color,
    #             marker=marker,
    #             xlabel=xlabel,
    #             ylabel=ylabel,
    #             **kwargs,
    #         )
    #         axes = plot.axes
    #     return plot

    # def _plot(self, x, **kwargs):
    #     plot = super().plot(f"{x}_v", f"{x}_s", yformat="0_ ", **kwargs)
    #     if fit:
    #         sigma = 2 if fit is True else fit
    #         column = self.add_column_for_fit(x, sigma)
    #         kwargs['marker'] = None
    #         kwargs['line'] = None
    #         kwargs.pop('axes')
    #         kwargs.pop('label')
    #         plot_ = super().plot(f'{x}_v', column, axes=plot.axes,
    #                              label=None, **kwargs)
    #         for series in plot_.series_collection:
    #             trendline = series.Trendlines().Add()
    #             plot_.axes.labels.append('__trendline__')
    #             trendline.DisplayEquation = True
    #             # trendline.Forward = 10
    #             # trendline.Backward = 10
    #         # print(plot_.axes.labels)
    #         # print(plot_.axes.legend.LegendEntries())
    #         # print(plot_.legend)
    #         plot_.axes.set_legend(**plot_.legend)
    #     return plot

    # def fit(self, x):
    #     pass

    # def add_column_for_fit(self, x, sigma):
    #     """

    #     Parameters
    #     ----------
    #     x : str
    #         変数名
    #     sigma: int or float
    #         フィッティングに用いるσ値の範囲
    #     """
    #     column_ = f"{x}_sf"
    #     if column_ in self.headers:
    #         return column_
    #     self[column_] = 1
    #     column = self.index_past(column_)
    #     sigma_cell = self.sheet.range(self.row - 1, column)
    #     sigma_cell.value = sigma
    #     set_font(sigma_cell, size=8, bold=True, italic=True, color="green")
    #     set_alignment(sigma_cell, "center")
    #     sigma = sigma_cell.get_address()
    #     row = self.cell.offset(self.columns.nlevels).row
    #     column_ref = self.index_past(f"{x}_s")
    #     cell_ref = self.sheet.range(row, column_ref)
    #     cell_ref = cell_ref.get_address(row_absolute=False)
    #     cell = self.sheet.range(row, column)
    #     range_ = self.sheet.range(cell, cell.offset(len(self) - 1))
    #     range_.api.NumberFormatLocal = "0.00_ "
    #     formula = f"=IF(AND({cell_ref}>=-{sigma},{cell_ref}<={sigma}),{cell_ref},NA())"
    #     range_.value = formula
    #     return column_


def select_index(index: Index, names: list[str]) -> Index:
    if not names:
        return Index(range(len(index)))

    if index.nlevels == 1:
        if index.name in names:
            return index
        return Index(range(len(index)))

    for name in index.names:
        if name not in names:
            index = index.droplevel(name)

    return index


def get_init_data(index: Index, columns: list[str]) -> DataFrame:
    columns = [f"{c}_{suffix}" for c, suffix in product(columns, ["n", "v", "s"])]
    array = np.zeros((len(index), len(columns)))
    return DataFrame(array, index=index, columns=columns)


def get_dist_func(dist: str | dict[str, str], columns: list[str]) -> dict[str, str]:
    if isinstance(dist, str):
        return {column: dist for column in columns}

    dist = dist.copy()
    for column in columns:
        dist.setdefault(column, "norm")

    return dist


def counter(cell: Range) -> str:
    start = cell.get_address()
    end = cell.get_address(row_absolute=False)
    return f"=AGGREGATE(3,1,{start}:{end})"


def sorted_value(parent_cell: Range, cell: Range, length: int) -> str:
    start = parent_cell.get_address()
    end = parent_cell.offset(length - 1).get_address()
    small = cell.get_address(row_absolute=False)
    return f"=IF({small}>0,AGGREGATE(15,1,{start}:{end},{small}),NA())"


def sigma_value(cell: Range, length: int, dist: str) -> str:
    small = cell.get_address(row_absolute=False)
    end = cell.offset(length - 1).get_address()

    if dist == "norm":
        return f"=IF({small}>0,NORM.S.INV({small}/({end}+1)),NA())"

    if dist == "weibull":
        return f"=IF({small}>0,LN(-LN(1-{small}/({end}+1))),NA())"

    msg = f"unknown distribution: {dist}"
    raise ValueError(msg)


def set_formula(cell: Range, length: int, formula: str) -> None:
    rng = Range((cell.row, cell.column), (cell.row + length - 1, cell.column))
    rng.value = formula



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Range as RangeImpl
from xlwings import Sheet

from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(
    scope="module",
    params=[(10, 10), (20, 10), (50, 10), (500, 500)],
    ids=lambda x: str(x),
)
def shape(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(scope="module")
def rng_impl(shape: tuple[int, int], sheet_module: Sheet):
    nrows, ncolumns = shape
    return sheet_module.range((1, 1), (nrows, ncolumns))


@pytest.fixture(scope="module")
def rng(shape: tuple[int, int], sheet_module: Sheet):
    nrows, ncolumns = shape
    return Range((1, 1), (nrows, ncolumns), sheet=sheet_module)


@pytest.mark.slow
def test_address_impl(benchmark, rng_impl: RangeImpl, rng: Range, shape):
    x = benchmark(lambda: [r.get_address() for r in rng_impl])
    assert len(x) == shape[0] * shape[1]
    assert x == list(rng.iter_addresses())


def test_address(benchmark, rng: Range, shape):
    x = benchmark(lambda: list(rng.iter_addresses()))
    assert len(x) == shape[0] * shape[1]


def test_address_frame(benchmark, rng: Range, shape):
    x = benchmark(lambda: rng.frame.get_address())
    assert x.shape == (shape[0], shape[1])


def test_address_iter_frame(benchmark, rng: Range, shape):
    def f():
        x = np.array(list(rng.iter_addresses())).reshape(shape[0], shape[1])
        return DataFrame(x)

    x = benchmark(f)
    assert x.shape == (shape[0], shape[1])



import pytest

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.dataframes.stats_frame import StatsFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def sf(sf_parent: SheetFrame):
    return StatsFrame(sf_parent, "count", by="x")


def test_len(sf: SheetFrame):
    assert len(sf) == 2


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["func", "x", "y", "z"]


@pytest.mark.parametrize(
    ("cell", "value"),
    [
        ("C3:C5", ["x", "a", "b"]),
        ("D3:D5", ["y", None, None]),
        ("E3:E5", ["z", None, None]),
        ("F3:F5", ["a", 10, 8]),
        ("G3:G5", ["b", 12, 8]),
        ("H3:H5", ["c", 10, 7]),
    ],
)
def test_value(sf: SheetFrame, cell, value):
    assert sf.sheet[cell].value == value



import pytest

from xlviews.core.index import Index, WideIndex


@pytest.fixture(scope="module")
def wide_index():
    return WideIndex({"A": [1, 2, 3], "B": [4, 5, 6, 7]})


def test_wide_getitem(wide_index: WideIndex):
    assert wide_index["A"] == [1, 2, 3]
    assert wide_index["B"] == [4, 5, 6, 7]


def test_wide_len(wide_index: WideIndex):
    assert len(wide_index) == 7


def test_wide_names(wide_index: WideIndex):
    assert wide_index.names == ["A", "B"]


def test_wide_to_list(wide_index: WideIndex):
    x = [("A", 1), ("A", 2), ("A", 3), ("B", 4), ("B", 5), ("B", 6), ("B", 7)]
    assert wide_index.to_list() == x


@pytest.mark.parametrize(("key", "loc"), [("A", (0, 3)), ("B", (3, 7))])
def test_wide_loc(wide_index: WideIndex, key, loc):
    assert wide_index.get_loc(key) == loc


def test_wide_append():
    index = WideIndex()
    index.append("A", [1, 2, 3])
    assert "A" in index


def test_wide_append_error():
    index = WideIndex({"A": [1, 2, 3]})
    with pytest.raises(ValueError, match="key 'A' already exists"):
        index.append("A", [1, 2, 3])


@pytest.fixture(scope="module")
def index(wide_index: WideIndex):
    return Index(["x", "y", "z"], wide_index)


def test_index_len(index: Index):
    assert len(index) == 10


def test_index_names(index: Index):
    assert index.names == [None]


def test_index_nlevels(index: Index):
    assert index.nlevels == 1


def test_index_to_list(index: Index):
    x = ["x", "y", "z"]
    y = [("A", 1), ("A", 2), ("A", 3), ("B", 4), ("B", 5), ("B", 6), ("B", 7)]
    assert index.to_list() == x + y


def test_index_append():
    index = Index(["a"])
    index.append("b")
    assert index.to_list() == ["a", "b"]


def test_index_append_wide():
    index = Index(["a"])
    index.append("b", [1, 2, 3])
    assert index.to_list() == ["a", ("b", 1), ("b", 2), ("b", 3)]



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import MultiIndexColumn

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return MultiIndexColumn(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_value_column(sf: SheetFrame):
    assert sf.cell.value is None
    v = sf.cell.offset(0, 3).expand().options(ndim=2).value
    assert len(v) == 10
    assert v[0] == ["a1", "a1", "a2", "a2"]
    assert v[1] == ["b1", "b2", "b1", "b2"]
    assert v[2] == [1, 11, 21, 31]
    assert v[-1] == [8, 18, 28, 38]


def test_value_values(sf: SheetFrame):
    v = sf.cell.offset(1, 0).expand().options(ndim=2).value
    assert len(v) == 9
    assert v[0] == ["x", "y", "z", "b1", "b2", "b1", "b2"]
    assert v[1] == [1, 1, 1, 1, 11, 21, 31]
    assert v[2] == [1, 1, 1, 2, 12, 22, 32]
    assert v[-1] == [2, 2, 1, 8, 18, 28, 38]


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.cell.get_address() == "$U$13"
    assert sf.row == 13
    assert sf.column == 21
    assert sf.sheet.name == sheet_module.name
    assert sf.index.nlevels == 3
    assert sf.columns.nlevels == 2


def test_len(sf: SheetFrame):
    assert len(sf) == 8


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["x", "y", "z"]


def test_contains(sf: SheetFrame):
    assert "x" not in sf
    assert "a1" in sf
    assert ("a1", "b1") not in sf


def test_iter(sf: SheetFrame):
    assert list(sf) == [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]


def test_value(sf: SheetFrame, df: DataFrame):
    df_sf = sf.value
    assert df_sf.equals(df.astype(float))
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)



import string

import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import MultiColumn

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return MultiColumn(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.cell.get_address() == "$K$2"
    assert sf.row == 2
    assert sf.column == 11
    assert sf.sheet.name == sheet_module.name
    assert sf.index.nlevels == 1
    assert sf.columns.nlevels == 4


def test_len(sf: SheetFrame):
    assert len(sf) == 6


def test_index_names(sf: SheetFrame):
    assert sf.index.names == [None]


def test_columns_names(sf: SheetFrame):
    assert sf.columns.names == ["s", "t", "r", "i"]


def test_iter(sf: SheetFrame):
    assert list(sf)[-1] == ("b", "d", 8, "y")


def test_value(sf: SheetFrame, df: DataFrame):
    df_sf = sf.value
    assert df_sf.equals(df.astype(float))
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)


@pytest.mark.parametrize(
    ("columns", "indexer"),
    [
        ({"s": "a", "t": "c"}, [12, 13, 14, 15]),
        ({"r": 4, "i": "x"}, [18]),
        ({"t": "c", "i": "x"}, [12, 14, 20, 22]),
    ],
)
def test_get_indexer(sf: SheetFrame, columns, indexer):
    assert all(sf.get_indexer(columns) == indexer)
    assert all(sf.get_indexer(**columns) == indexer)


def test_iter_ranges(sf: SheetFrame):
    for rng, i in zip(sf.iter_ranges(), range(11, 26), strict=False):
        c = string.ascii_uppercase[i]
        assert rng.get_address() == f"${c}$6:${c}$11"


@pytest.fixture(scope="module")
def df_melt(sf: SheetFrame):
    return sf.melt(formula=True, value_name="v")


def test_melt_len(df_melt: DataFrame):
    assert len(df_melt) == 16


def test_melt_columns(df_melt: DataFrame):
    assert df_melt.columns.to_list() == ["s", "t", "r", "i", "v"]


@pytest.mark.parametrize(
    ("i", "v"),
    [
        (0, ["a", "c", 1, "x", "=$L$6:$L$11"]),
        (1, ["a", "c", 1, "y", "=$M$6:$M$11"]),
        (2, ["a", "c", 2, "x", "=$N$6:$N$11"]),
        (7, ["a", "d", 4, "y", "=$S$6:$S$11"]),
        (14, ["b", "d", 8, "x", "=$Z$6:$Z$11"]),
        (15, ["b", "d", 8, "y", "=$AA$6:$AA$11"]),
    ],
)
def test_melt_value(df_melt: DataFrame, i, v):
    assert df_melt.iloc[i].to_list() == v


def test_agg(sf: SheetFrame):
    df_sf = sf.agg()
    df_melt = sf.melt()
    assert isinstance(df_sf, DataFrame)
    assert df_sf.equals(df_melt)
    assert df_sf.index.equals(df_melt.index)
    assert df_sf.columns.equals(df_melt.columns)


def test_number_format(sf: SheetFrame):
    sf.number_format("0.000", s="a", r=3, autofit=True)
    i = sf.get_indexer({"s": "a", "r": 3})
    assert sf.sheet.range(6, i[0]).number_format == "0.000"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from numpy.typing import NDArray
    from pandas import DataFrame
    from pandas._typing import Axes


class WideIndex(dict[str, list[Any]]):
    """Represent a wide index."""

    def __len__(self) -> int:
        return sum(len(values) for values in self.values())

    @property
    def names(self) -> list[str]:
        return list(self.keys())

    def to_list(self) -> list[tuple[str, Any]]:
        return [(key, value) for key in self for value in self[key]]

    def get_loc(self, key: str) -> tuple[int, int]:
        start = [k for k, _ in self.to_list()].index(key)
        stop = start + len(self[key])
        return start, stop

    def append(self, key: str, values: Iterable[Any]) -> None:
        if key in self:
            msg = f"key {key!r} already exists"
            raise ValueError(msg)

        self[key] = list(values)


class Index:
    index: pd.Index
    wide_index: WideIndex

    def __init__(
        self,
        index: Axes,
        wide_index: WideIndex | dict[str, Any] | None = None,
    ) -> None:
        self.index = index if isinstance(index, pd.Index) else pd.Index(index)

        if wide_index is None:
            wide_index = WideIndex()
        elif not isinstance(wide_index, WideIndex):
            wide_index = WideIndex(wide_index)

        self.wide_index = wide_index

    def __len__(self) -> int:
        return len(self.index) + len(self.wide_index)

    @property
    def names(self) -> list[str]:
        return self.index.names

    @property
    def nlevels(self) -> int:
        """Get the number of levels in the index.

        Examples:
            >>> index = Index(["a"], {"b": [1, 2, 3]})
            >>> index.nlevels
            1

            >>> index = Index(pd.MultiIndex.from_tuples([("a", 1), ("a", 2)]))
            >>> index.nlevels
            2
        """
        return self.index.nlevels

    def to_list(self) -> list[Any]:
        """Convert the index to a list.

        Examples:
            >>> index = Index(["a"], {"b": [1, 2, 3]})
            >>> index.to_list()
            ['a', ('b', 1), ('b', 2), ('b', 3)]
        """
        return [*self.index.to_list(), *self.wide_index.to_list()]

    def __iter__(self) -> Iterator[str]:
        """Iterate over the index.

        Examples:
            >>> index = Index(["a"], {"b": [1, 2, 3]})
            >>> list(index)
            ['a', 'b']
        """
        yield from self.index.to_list()
        yield from self.wide_index

    def __contains__(self, key: Any) -> bool:
        """Check if the index contains a key.

        Examples:
            >>> index = Index(["a"], {"b": [1, 2, 3]})
            >>> "a" in index
            True
            >>> "b" in index
            True
            >>> "c" in index
            False
            >>> ("b", 1) in index
            False
        """
        if not isinstance(key, str):
            return False

        return any(key in k for k in self)

    def append(self, key: str, values: Iterable[Any] | None = None) -> None:
        """Append a key to the index.

        Args:
            key (str): The key to append.
            values (Iterable[Any], optional): The values to append.

        Examples:
            >>> index = Index(["a"])
            >>> index.append("b")
            >>> index.to_list()
            ['a', 'b']

            >>> index.append("c", [1, 2, 3])
            >>> index.to_list()
            ['a', 'b', ('c', 1), ('c', 2), ('c', 3)]
        """
        if values is None:
            self.index = self.index.append(pd.Index([key]))
        else:
            self.wide_index.append(key, values)

    def get_loc(self, key: str, offset: int = 0) -> int | tuple[int, int]:
        """Get the location of a key. The end index is inclusive.

        Examples:
            >>> index = Index(["a", "b"], {"c": [1, 2, 3], "d": [4, 5]})
            >>> index.get_loc("a")
            0
            >>> index.get_loc("b", offset=10)
            11
            >>> index.get_loc("c")
            (2, 4)
            >>> index.get_loc("d", offset=20)
            (25, 26)
        """
        if key not in self.index:
            offset = len(self.index) + offset
            loc = self.wide_index.get_loc(key)
            return loc[0] + offset, loc[1] + offset - 1

        loc = self.index.get_loc(key)
        if isinstance(loc, int):
            return loc + offset

        raise NotImplementedError

    def get_indexer(
        self,
        columns: dict[str, Any] | None = None,
        offset: int = 0,
        **kwargs,
    ) -> NDArray[np.intp]:
        if self.index.nlevels == 1:
            raise NotImplementedError

        if columns is not None:
            kwargs.update(columns)

        idx = [self.index.get_level_values(k) == v for k, v in kwargs.items()]

        return np.where(np.all(idx, axis=0))[0] + offset

    def to_frame(self, index: bool = True) -> DataFrame:
        return self.index.to_frame(index=index)



"""Set styles for SheetFrame."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import pywintypes
from xlwings.constants import TableStyleElementType

from xlviews.config import rcParams
from xlviews.style import (
    EVEN_COLOR,
    ODD_COLOR,
    hide_succession,
    hide_unique,
    set_alignment,
    set_banding,
    set_border,
    set_fill,
    set_font,
)
from xlviews.utils import iter_group_locs, suspend_screen_updates

if TYPE_CHECKING:
    from pandas import Index
    from xlwings import Range, Sheet

    from xlviews.colors import Color

    from .heat_frame import HeatFrame
    from .sheet_frame import SheetFrame
    from .table import Table


def _set_style(
    start: Range,
    end: Range | None,
    name: str,
    *,
    border: bool = True,
    fill: bool = True,
    font: bool = True,
    font_size: int | None = None,
) -> None:
    rng = start.sheet.range(start, end)

    if border:
        edge_color = rcParams["frame.border.color"]
        inside_color = rcParams["frame.border.inside.color"]
        set_border(rng, edge_color=edge_color, inside_color=inside_color)

    if fill:
        set_fill(rng, color=rcParams[f"frame.{name}.fill.color"])

    if font:
        color = rcParams[f"frame.{name}.font.color"]
        bold = rcParams[f"frame.{name}.font.bold"]
        size = font_size or rcParams["frame.font.size"]
        set_font(rng, color=color, bold=bold, size=size)


@suspend_screen_updates
def set_frame_style(
    sf: SheetFrame,
    *,
    border: bool = True,
    fill: bool = True,
    font: bool = True,
    font_size: int | None = None,
    alignment: str | None = "center",
    banding: bool = False,
    succession: bool = False,
) -> None:
    """Set style of SheetFrame.

    Args:
        sf: The SheetFrame object.
        border: Whether to draw the border.
        font: Whether to specify the font.
        fill: Whether to fill the frame.
        font_size: The font size to specify directly.
        alignment: The alignment of the frame.
        banding: Whether to draw the banding.
        succession: Whether to hide the succession of the index.
    """
    cell = sf.cell
    sheet = sf.sheet

    set_style = partial(
        _set_style,
        border=border,
        fill=fill,
        font=font,
        font_size=font_size,
    )

    index_nlevels = sf.index.nlevels
    columns_nlevels = sf.columns.nlevels
    length = len(sf)

    end = cell.offset(columns_nlevels - 1, index_nlevels - 1)
    if columns_nlevels > 1 and index_nlevels == 1:
        set_style(cell, end, "columns.name")
    else:
        set_style(cell, end, "index.name")

    start = cell.offset(columns_nlevels, 0)
    end = cell.offset(columns_nlevels + length - 1, index_nlevels - 1)
    set_style(start, end, "index")

    if succession:
        rng = sheet.range(start.offset(1, 0), end)
        hide_succession(rng)

        start = cell.offset(columns_nlevels - 1, 0)
        end = cell.offset(columns_nlevels - 1, index_nlevels - 1)
        rng = sheet.range(start, end)
        hide_unique(rng, length)

    width = len(sf.columns)

    start = cell.offset(0, index_nlevels)
    end = cell.offset(columns_nlevels - 1, index_nlevels + width - 1)
    set_style(start, end, "columns")

    start = cell.offset(columns_nlevels, index_nlevels)
    end = cell.offset(columns_nlevels + length - 1, index_nlevels + width - 1)
    set_style(start, end, "values")

    rng = sheet.range(start, end)

    if banding:
        set_banding(rng)

    rng = sheet.range(cell, end)

    if border:
        ew = rcParams["frame.border.weight"]
        ec = rcParams["frame.border.color"]
        set_border(rng, edge_weight=ew, inside_weight=0, edge_color=ec)

    if alignment:
        set_alignment(rng, alignment)


def set_wide_column_style(sf: SheetFrame) -> None:
    edge_color = rcParams["frame.border.color"]
    edge_weight = rcParams["frame.wide-columns.border.weight"]

    columns = list(sf.columns.wide_index)
    for column in columns:
        loc = sf.columns.get_loc(column, sf.column + sf.index.nlevels)
        if not isinstance(loc, tuple):
            raise NotImplementedError

        rng = sf.sheet.range((sf.row, loc[0]), (sf.row, loc[1]))

        er = edge_weight if column == columns[-1] else 2
        edge_weight_tuple = (2, er, 2, 2)
        set_border(rng, edge_weight_tuple, inside_weight=1, edge_color=edge_color)
        _set_style(rng, None, "wide-columns", border=False)

        rng = sf.sheet.range((sf.row - 1, loc[0]), (sf.row - 1, loc[1]))

        el = edge_weight if column == columns[0] else 2
        edge_weight_tuple = (el, edge_weight, edge_weight, 2)
        set_border(rng, edge_weight_tuple, inside_weight=0, edge_color=edge_color)
        _set_style(rng, None, "wide-columns.name", border=False)


def set_table_style(
    table: Table,
    even_color: Color = EVEN_COLOR,
    odd_color: Color = ODD_COLOR,
) -> None:
    book = table.sheet.book.api

    try:
        style = book.TableStyles("xlviews")
    except pywintypes.com_error:
        style = book.TableStyles.Add("xlviews")
        odd_type = TableStyleElementType.xlRowStripe1
        style.TableStyleElements(odd_type).Interior.Color = odd_color
        even_type = TableStyleElementType.xlRowStripe2
        style.TableStyleElements(even_type).Interior.Color = even_color

    table.api.TableStyle = style


@suspend_screen_updates
def set_heat_frame_style(
    sf: HeatFrame,
    *,
    border: bool = True,
    fill: bool = True,
    font: bool = True,
    font_size: int | None = None,
    alignment: str | None = "center",
) -> None:
    """Set style of SheetFrame.

    Args:
        sf: The SheetFrame object.
        border: Whether to draw the border.
        fill: Whether to fill the frame.
        font: Whether to specify the font.
        font_size: The font size to specify directly.
        alignment: The alignment of the frame.
    """
    cell = sf.cell
    sheet = sf.sheet

    set_style = partial(
        _set_style,
        border=border,
        fill=fill,
        font=font,
        font_size=font_size,
    )

    index_nlevels = sf.index.nlevels
    columns_nlevels = sf.columns.nlevels
    length = len(sf)

    start = cell.offset(columns_nlevels, 0)
    end = cell.offset(columns_nlevels + length - 1, index_nlevels - 1)
    set_style(start, end, "index")

    width = len(sf.columns)

    start = cell.offset(columns_nlevels - 1, index_nlevels)
    end = cell.offset(columns_nlevels - 1, index_nlevels + width - 1)
    set_style(start, end, "index")

    start = cell.offset(columns_nlevels, index_nlevels)
    end = cell.offset(columns_nlevels + length - 1, index_nlevels + width - 1)
    set_style(start, end, "values")

    rng = sheet.range(cell, end)

    if alignment:
        set_alignment(rng, alignment)

    _merge_index(sf.columns, sf.row, sf.column, 1, sf.sheet)
    _merge_index(sf.index, sf.row, sf.column, 0, sf.sheet)
    _set_heat_border(sf)


def _merge_index(index: Index, row: int, column: int, axis: int, sheet: Sheet) -> None:
    for start, end in iter_group_locs(index):
        if start == end:
            continue
        if axis == 0:
            sheet.range((row + start + 1, column), (row + end + 1, column)).merge()
        else:
            sheet.range((row, column + start + 1), (row, column + end + 1)).merge()


def _set_heat_border(sf: HeatFrame) -> None:
    r = sf.row + 1
    c = sf.column + 1

    ec = rcParams["heat.border.color"]

    for row in iter_group_locs(sf.index, offset=r):
        for col in iter_group_locs(sf.columns, offset=c):
            if row[0] == row[1] and col[0] == col[1]:
                continue

            rng = sf.sheet.range((row[0], col[0]), (row[1], col[1]))
            set_border(rng, edge_weight=2, edge_color=ec, inside_weight=0)



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType, LineStyle, MarkerStyle

from xlviews.chart.axes import Axes
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def x(sheet_module: Sheet):
    x = sheet_module["B2:B11"]
    x.options(transpose=True).value = list(range(10))
    return x


@pytest.fixture(scope="module")
def y(sheet_module: Sheet):
    y = sheet_module["C2:C11"]
    y.options(transpose=True).value = list(range(10, 20))
    return y


@pytest.fixture(scope="module")
def ax(sheet_module: Sheet):
    ct = ChartType.xlXYScatterLines
    return Axes(300, 10, chart_type=ct, sheet=sheet_module)


@pytest.mark.parametrize(
    ("style", "value", "size"),
    [
        ("", MarkerStyle.xlMarkerStyleNone, 5),
        ("o", MarkerStyle.xlMarkerStyleCircle, 10),
        ("^", MarkerStyle.xlMarkerStyleTriangle, 9),
        ("s", MarkerStyle.xlMarkerStyleSquare, 8),
        ("d", MarkerStyle.xlMarkerStyleDiamond, 7),
        ("+", MarkerStyle.xlMarkerStylePlus, 6),
        ("x", MarkerStyle.xlMarkerStyleX, 5),
        (".", MarkerStyle.xlMarkerStyleDot, 4),
        ("-", MarkerStyle.xlMarkerStyleDash, 3),
        ("*", MarkerStyle.xlMarkerStyleStar, 2),
    ],
)
def test_series_style_marker(ax: Axes, x, y, style, value, size):
    series = ax.add_series(x, y, label="a")
    series.marker(style, size=size)
    assert series.api.MarkerStyle == value
    assert series.api.MarkerSize == size
    series.delete()


@pytest.mark.parametrize(
    ("style", "value", "weight"),
    [
        ("", LineStyle.xlLineStyleNone, 2),
        ("-", LineStyle.xlContinuous, 1),
        ("--", LineStyle.xlDash, 2),
        ("-.", LineStyle.xlDashDot, 1),
        (".", LineStyle.xlDot, 2),
    ],
)
def test_series_style_line(ax: Axes, x, y, style, value, weight):
    series = ax.add_series(x, y, label="a")
    series.line(style, weight=weight)
    assert series.api.Border.LineStyle == value
    series.delete()


def test_series_style_line_none(x, y, sheet: Sheet):
    ct = ChartType.xlXYScatterLines
    ax = Axes(300, 10, chart_type=ct, sheet=sheet)
    series = ax.add_series(x, y, label="a")
    style = series.api.Border.LineStyle
    series.line()
    assert series.api.Border.LineStyle == style


@pytest.mark.parametrize(
    ("color", "value", "alpha"),
    [
        ("red", 255, 0.5),
        ("green", 32768, 0.2),
        ("blue", 16711680, 0.4),
        ("lime", 65280, 0.7),
    ],
)
def test_series_style_color(ax: Axes, x, y, color, value, alpha):
    series = ax.add_series(x, y, label="a")
    series.line("-", color=color, alpha=alpha)
    assert value == series.api.Format.Line.ForeColor.RGB
    series.delete()



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.dataframes.heat_frame import HeatFrame

if TYPE_CHECKING:
    from pandas import DataFrame

    from xlviews.dataframes.sheet_frame import SheetFrame


class HeatFrameContainer:
    row: int = 0
    column: int = 0
    df: DataFrame
    sf: HeatFrame

    def __init__(self, sf: SheetFrame, row: int = 0, column: int = 0) -> None:
        self.row = row or self.row or sf.row
        self.column = column or self.column or (sf.column + sf.width + 1)
        self.df = self.dataframe(sf)
        self.sf = HeatFrame(self.row, self.column, data=self.df, sheet=sf.sheet)
        self.init()

    def init(self) -> None:
        self.sf.autofit()

    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        raise NotImplementedError



import numpy as np
import pytest
from pandas import Index, MultiIndex
from scipy.stats import norm

from xlviews.dataframes.dist_frame import DistFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


def test_select_index():
    from xlviews.dataframes.dist_frame import select_index

    index = Index(["a", "b"], name="x")
    x = select_index(index, ["x"])
    assert x.equals(index)


def test_select_index_none():
    from xlviews.dataframes.dist_frame import select_index

    index = Index(["a", "b"], name="x")
    x = select_index(index, ["y"])
    assert x.equals(Index([0, 1]))


def test_select_index_multi_one():
    from xlviews.dataframes.dist_frame import select_index

    index = MultiIndex.from_tuples([(1, 2), (3, 4)], names=["x", "y"])
    x = select_index(index, ["x"])
    assert isinstance(x, Index)
    assert not isinstance(x, MultiIndex)
    assert x.name == "x"
    assert x.names == ["x"]


def test_select_index_multi_two():
    from xlviews.dataframes.dist_frame import select_index

    index = MultiIndex.from_tuples([(1, 2), (3, 4)], names=["x", "y"])
    x = select_index(index, ["x", "y"])
    assert isinstance(x, MultiIndex)
    assert x.names == ["x", "y"]


def test_select_index_multi_none():
    from xlviews.dataframes.dist_frame import select_index

    index = MultiIndex.from_tuples([(1, 2), (3, 4)], names=["x", "y"])
    x = select_index(index, [])
    assert x.equals(Index([0, 1]))


def test_init_data(sf_parent: SheetFrame):
    from xlviews.dataframes.dist_frame import get_init_data

    df = get_init_data(sf_parent.index, ["a", "b"])
    c = ["a_n", "a_v", "a_s", "b_n", "b_v", "b_s"]
    assert df.columns.to_list() == c
    assert df.index.names == ["x", "y"]
    assert len(df) == 14


@pytest.mark.parametrize(
    ("cell", "value"),
    [
        ("G4", 1),
        ("I4", 1),
        ("J4", 1),
        ("I7", 4),
        ("J7", 4),
        ("I17", 2),
        ("J17", 2),
        ("I13", 1),
        ("I14", 2),
        ("I15", 3),
        ("L13", 1),
        ("L14", 2),
        ("L15", 2),
        ("K4", norm.ppf(1 / 6)),
        ("N5", norm.ppf(2 / 6)),
        ("K6", norm.ppf(3 / 6)),
        ("N7", norm.ppf(4 / 6)),
        ("K8", norm.ppf(5 / 6)),
        ("K13", norm.ppf(1 / 4)),
        ("K14", norm.ppf(2 / 4)),
        ("K15", norm.ppf(3 / 4)),
        ("N13", norm.ppf(1 / 3)),
        ("N14", norm.ppf(2 / 3)),
        ("N15", norm.ppf(2 / 3)),
        ("N16", norm.ppf(1 / 3)),
        ("K17", norm.ppf(2 / 3)),
    ],
)
def test_value(sf: DistFrame, cell: str, value: float):
    v = sf.sheet[cell].value
    assert v is not None
    np.testing.assert_allclose(v, value)



from __future__ import annotations

from pandas import DataFrame
from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.testing.common import FrameContainer, create_sheet


class Base(FrameContainer):
    @classmethod
    def dataframe(cls) -> DataFrame:
        a = ["c"] * 10
        b = ["s"] * 5 + ["t"] * 5
        c = ([100] * 2 + [200] * 3) * 2
        x = list(range(10))
        y = list(range(10, 20))
        df = DataFrame({"a": a, "b": b, "c": c, "x": x, "y": y})
        return df.set_index(["a", "b", "c"])


if __name__ == "__main__":
    sheet = create_sheet()
    fc = Base(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)

    data = sf.agg(include_sheetname=True)
    ax = Axes(2, 8, chart_type=ChartType.xlXYScatter)
    s = ax.add_series(data["x"], data["y"])
    s.set("o", color="red", alpha=0.6)

    ax.set(
        xlabel="xlabel",
        ylabel="ylabel",
        title="title",
        xticks=(0, 10, 2),
        legend=(1, -1),
    )

    ax = Axes(chart_type=ChartType.xlXYScatterLines)
    df = sf.groupby("b").agg(include_sheetname=True)
    for key, s in df.iterrows():
        x = ax.add_series(s["x"], s["y"], label=f"{key}")
        x.set("--", color="red", marker="o")

    ax.set(
        xlabel="xlabel",
        ylabel="ylabel",
        title="title",
        xticks=(0, 10, 5),
        legend=(-1, -1),
    )

    ax = Axes(chart_type=ChartType.xlXYScatterLinesNoMarkers)
    df = sf.groupby(["b", "c"]).agg(include_sheetname=True)
    for key, s in df.iterrows():
        x = ax.add_series(s["x"], s["y"], label=f"{key[0]}_{key[1]}")  # type: ignore
        x.set(marker="o")

    ax.set(
        xlabel="xlabel",
        ylabel="ylabel",
        title="title",
        xticks=(0, 20, 4),
        yticks=(0, 20, 4),
        legend=(1, 1),
    )



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available


def test_func_none():
    from xlviews.dataframes.stats_frame import get_func

    func = ["count", "max", "mean", "median", "min", "soa"]
    assert sorted(get_func(None)) == func


def test_func_str():
    from xlviews.dataframes.stats_frame import get_func

    assert get_func("count") == ["count"]


@pytest.mark.parametrize("func", [["count"], {"a": "count"}])
def test_func_else(func):
    from xlviews.dataframes.stats_frame import get_func

    assert get_func(func) == func


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
@pytest.mark.parametrize(
    ("funcs", "n"),
    [(["mean"], 4), (["min", "max", "median"], 12)],
)
def test_length(sf_parent: SheetFrame, funcs, n):
    from xlviews.dataframes.stats_frame import get_length

    assert get_length(sf_parent, ["x", "y"], funcs) == n


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_length_none_list(sf_parent: SheetFrame):
    from xlviews.dataframes.stats_frame import get_length

    assert get_length(sf_parent, [], ["min", "max"]) == 2


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_has_header(sf_parent: SheetFrame):
    from xlviews.dataframes.stats_frame import has_header

    assert has_header(sf_parent)


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_move_down(sheet: Sheet):
    from xlviews.dataframes.stats_frame import move_down

    df = DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])
    sf = SheetFrame(3, 3, data=df, sheet=sheet)
    assert sheet["D3:F3"].value == ["a", "b", "c"]
    assert sheet["D2:F2"].value == [None, None, None]
    assert move_down(sf, 3) == 3
    assert sheet["D6:F6"].value == ["a", "b", "c"]
    assert sheet["D5:F5"].value == [None, None, None]
    assert sf.row == 6


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_move_down_header(sheet: Sheet):
    from xlviews.dataframes.stats_frame import move_down

    df = DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])
    sf = SheetFrame(3, 3, data=df, sheet=sheet)
    sheet["D2"].value = "x"
    assert sheet["D3:F3"].value == ["a", "b", "c"]
    assert sheet["D2:F2"].value == ["x", None, None]
    assert move_down(sf, 3) == 4
    assert sheet["D7:F7"].value == ["a", "b", "c"]
    assert sheet["D6:F6"].value == ["x", None, None]
    assert sf.row == 7



from __future__ import annotations

from collections.abc import Callable, Hashable
from itertools import product
from typing import TYPE_CHECKING, TypeAlias

import pandas as pd
from pandas import DataFrame, Index

from .palette import PaletteStyle, get_color_palette, get_marker_palette

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any, Self

    from xlviews.chart.axes import Axes
    from xlviews.chart.series import Series

Label: TypeAlias = str | Callable[[dict[str, Hashable]], str]


class Plot:
    axes: Axes
    data: DataFrame
    index: list[tuple[Hashable, ...]]
    series_collection: list[Series]

    def __init__(self, axes: Axes, data: DataFrame | pd.Series) -> None:
        self.axes = axes

        if isinstance(data, pd.Series):
            data = data.to_frame().T

        self.data = data

    def add(
        self,
        x: str | list[str],
        y: str | list[str],
        chart_type: int | None = None,
    ) -> Self:
        self.index = []
        self.series_collection = []

        xs = x if isinstance(x, list) else [x]
        ys = y if isinstance(y, list) else [y]

        for x, y in product(xs, ys):
            for idx, s in self.data.iterrows():
                series = self.axes.add_series(s[x], s[y], chart_type=chart_type)
                index = idx if isinstance(idx, tuple) else (idx,)
                self.index.append(index)
                self.series_collection.append(series)

        return self

    def keys(self) -> Iterator[dict[str, Hashable]]:
        names = self.data.index.names

        for index in self.index:
            yield dict(zip(names, index, strict=True))

    def set(
        self,
        label: Label | None = None,
        marker: PaletteStyle | None = None,
        color: PaletteStyle | None = None,
        alpha: float | None = None,
        weight: float | None = None,
        size: int | None = None,
    ) -> Self:
        marker_palette = get_marker_palette(self.data, marker)
        color_palette = get_color_palette(self.data, color)

        for key, s in zip(self.keys(), self.series_collection, strict=True):
            s.set(
                label=label and get_label(label, key),
                color=color_palette and color_palette[key],
                marker=marker_palette and marker_palette[key],
                alpha=alpha,
                weight=weight,
                size=size,
            )

        return self

    @classmethod
    def facet(
        cls,
        axes: Axes,
        data: DataFrame,
        index: str | list[str] | None = None,
        columns: str | list[str] | None = None,
    ) -> Iterator[tuple[dict[str, Any], Self]]:
        left = axes.chart.left
        top = axes.chart.top
        width = axes.chart.width
        height = axes.chart.height

        for r, rkey in enumerate(iterrows(data.index, index)):
            for c, ckey in enumerate(iterrows(data.index, columns)):
                key = rkey | ckey
                sub = xs(data, key)

                if len(sub) == 0:
                    continue

                if r == 0 and c == 0:
                    axes_ = axes
                else:
                    axes_ = axes.copy(left=left + c * width, top=top + r * height)

                yield key, cls(axes_, sub)


def get_label(label: Label, key: dict[str, Hashable]) -> str:
    if isinstance(label, str):
        return label.format(**key)

    if callable(label):
        return label(key)

    msg = f"Invalid label: {label}"
    raise ValueError(msg)


def iterrows(
    index: Index,
    levels: int | str | Sequence[int | str] | None,
) -> Iterator[dict[str, Any]]:
    if levels is None:
        yield {}
        return

    if isinstance(levels, int | str):
        levels = [levels]

    if levels:
        values = {level: index.get_level_values(level) for level in levels}
        it = DataFrame(values).drop_duplicates().iterrows()

        for _, s in it:
            yield s.to_dict()


def xs(df: DataFrame, index: dict[str, Any] | None) -> DataFrame:
    if index:
        df = df.xs(tuple(index.values()), 0, tuple(index.keys()), drop_level=False)  # type: ignore

    return df



import numpy as np
import pytest
from pandas import DataFrame, Series
from xlwings import Sheet

from xlviews.dataframes.groupby import GroupBy, to_dict
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    ("keys", "values", "expected"),
    [
        ([1, 2, 1], ["a", "b", "c"], {1: ["a", "c"], 2: ["b"]}),
        (["x", "y", "x"], [10, 20, 30], {"x": [10, 30], "y": [20]}),
        ([True, False, True], [1.1, 2.2, 3.3], {True: [1.1, 3.3], False: [2.2]}),
        ([], [], {}),
        ([None, None], ["a", "b"], {None: ["a", "b"]}),
    ],
)
def test_to_dict(keys, values, expected):
    assert to_dict(keys, values) == expected


@pytest.mark.parametrize("func", [lambda x: x, np.array, Series])
def test_create_group_index_series(func):
    from xlviews.dataframes.groupby import create_group_index

    values = [1, 1, 2, 2, 2, 3, 3, 1, 1, 2, 2, 3, 3]
    index = create_group_index(func(values))
    assert index[(1,)] == [(0, 1), (7, 8)]
    assert index[(2,)] == [(2, 4), (9, 10)]
    assert index[(3,)] == [(5, 6), (11, 12)]


@pytest.mark.parametrize("func", [lambda x: x, DataFrame])
def test_create_group_index_dataframe(func):
    from xlviews.dataframes.groupby import create_group_index

    values = [[1, 2], [1, 2], [3, 4], [3, 4], [1, 2], [3, 4], [3, 4]]
    index = create_group_index(func(values))
    assert index[(1, 2)] == [(0, 1), (4, 4)]
    assert index[(3, 4)] == [(2, 3), (5, 6)]


@pytest.fixture(scope="module")
def sf(sheet_module: Sheet):
    a = ["c"] * 10
    b = ["s"] * 5 + ["t"] * 5
    c = ([100] * 2 + [200] * 3) * 2
    x = list(range(10))
    y = list(range(10, 20))
    df = DataFrame({"a": a, "b": b, "c": c, "x": x, "y": y})
    df = df.set_index(["a", "b", "c"])
    return SheetFrame(2, 2, df, sheet_module)


@pytest.mark.parametrize(
    ("by", "n"),
    [
        ("a", 1),
        ("b", 2),
        ("c", 2),
        (["a", "b"], 2),
        (["a", "c"], 2),
        (["b", "c"], 4),
        (["a", "b", "c"], 4),
    ],
)
def test_len(sf: SheetFrame, by, n: int):
    gr = GroupBy(sf, by)
    assert len(gr) == n


@pytest.fixture(scope="module")
def gp(sf: SheetFrame):
    return GroupBy(sf, ["a", "c"])


def test_keys(gp: GroupBy):
    keys = [("c", 100), ("c", 200)]
    assert list(gp.keys()) == keys


def test_values(gp: GroupBy):
    values = [[(3, 4), (8, 9)], [(5, 7), (10, 12)]]
    assert list(gp.values()) == values


def test_items(gp: GroupBy):
    assert next(gp.items()) == (("c", 100), [(3, 4), (8, 9)])


def test_iter(gp: GroupBy):
    assert next(iter(gp)) == ("c", 100)


@pytest.mark.parametrize(
    ("key", "value"),
    [
        (("c", 100), [(3, 4), (8, 9)]),
        (("c", 200), [(5, 7), (10, 12)]),
    ],
)
def test_getitem(gp: GroupBy, key, value):
    assert gp[key] == value


@pytest.fixture
def sf2(sheet: Sheet):
    df = DataFrame(
        {
            "x": ["a"] * 8 + ["b"] * 8 + ["a"] * 4,
            "y": (["c"] * 4 + ["d"] * 4) * 2 + ["c"] * 4,
            "z": range(1, 21),
            "a": range(1, 21),
        },
    )
    df = df.set_index(["x", "y", "z"])
    return SheetFrame(3, 3, data=df, sheet=sheet)


@pytest.mark.parametrize(
    ("by", "n"),
    [(None, 1), ("x", 2), (["x", "y"], 4), (["x", "y", "z"], 20)],
)
def test_by(sf2: SheetFrame, by, n):
    gp = GroupBy(sf2, by)
    assert len(gp.group) == n


def test_group_key(sf2: SheetFrame):
    gp = GroupBy(sf2, ["x", "y"])
    keys = list(gp.group.keys())
    assert keys == [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")]



"""Read the configuration file.

The configuration parameters are described in `xlviews.toml`
located in the same directory.

Custom settings can be made by directly modifying `xlviews.config.rcParams`.
"""

from __future__ import annotations

import tomllib
from pathlib import Path
from typing import Any

CONFIG_FILE = Path(__file__).parent / "xlviews.toml"


def load_config() -> dict[str, Any]:
    with CONFIG_FILE.open("rb") as f:
        return tomllib.load(f)


class Config:
    params: dict[str, Any]

    def __init__(self) -> None:
        self.params = load_config()

    def __getitem__(self, key: str) -> Any:
        keys = key.split(".")
        value = self.params

        for k in keys:
            value = value[k]

        return value

    def __setitem__(self, key: str, value: Any) -> None:
        keys = key.split(".")
        params = self.params

        for k in keys[:-1]:
            params = params[k]

        params[keys[-1]] = value

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default


rcParams = Config()  # noqa: N816



from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias

from xlwings import Range as RangeImpl

from .range import Range
from .range_collection import RangeCollection

if TYPE_CHECKING:
    from collections.abc import Iterable

Func: TypeAlias = str | Range | RangeImpl | None

NONCONST_VALUE = "*"


def const(rng: Range | RangeImpl) -> str:
    """Return a formula to check if the values in the range are unique."""
    column = rng.get_address(column_absolute=False)
    ref = rng[0].offset(-1).get_address(column_absolute=False)

    subtotal = f"SUBTOTAL(3,{column})"
    name = f'SUBSTITUTE(ADDRESS(ROW({column}),COLUMN({column}),4),ROW({column}),"")'
    index = f"INDEX(SUBTOTAL(3,INDIRECT({name}&ROW({column}))),)"
    value = f"INDEX({column},MATCH(1,{index},0))"

    prod_first = f'SUBTOTAL(3,OFFSET({ref},ROW(INDIRECT("1:"&ROWS({column}))),))'
    prod_second = f"({column}={value})"
    sumproduct = f"SUMPRODUCT({prod_first}*{prod_second})"

    return f'IFNA(IF({subtotal}={sumproduct},{value},"{NONCONST_VALUE}"),"")'


AGG_FUNCS = {
    "median": 12,
    "soa": 999,
    "count": 2,
    "min": 5,
    "mean": 1,
    "max": 4,
    "std": 8,
    "sum": 9,
}

AGG_FUNCS_SORTED = {key: AGG_FUNCS[key] for key in sorted(AGG_FUNCS.keys())}
AGG_FUNC_NAMES = ",".join(f'"{name}"' for name in AGG_FUNCS_SORTED)
AGG_FUNC_INTS = ",".join(f'"{value}"' for value in AGG_FUNCS_SORTED.values())


def _aggregate(
    func: Func,
    ranges: Range | RangeCollection | Iterable[Range | RangeCollection] | str,
    option: int,
    **kwargs,
) -> str:
    if func == "soa":
        std = aggregate("std", ranges, option, **kwargs)
        median = aggregate("median", ranges, option, **kwargs)
        return f"{std}/{median}"

    if isinstance(ranges, Range | RangeCollection):
        ranges = [ranges]

    if isinstance(ranges, str):
        column = ranges
    else:
        column = ",".join(r.get_address(**kwargs) for r in ranges)

    if func is None:
        return column

    if isinstance(func, str):
        if func in AGG_FUNCS:
            return f"AGGREGATE({AGG_FUNCS[func]},{option},{column})"

        msg = f"Invalid aggregate function: {func}"
        raise ValueError(msg)

    ref = func.get_address(column_absolute=False, row_absolute=False)
    func = f"LOOKUP({ref},{{{AGG_FUNC_NAMES}}},{{{AGG_FUNC_INTS}}})"
    soa = aggregate("soa", ranges, option=option, **kwargs)
    return f'IF({ref}="soa",{soa},AGGREGATE({func},{option},{column}))'


def aggregate(
    func: Func,
    ranges: Range | RangeCollection | Iterable[Range | RangeCollection] | str,
    option: int = 7,  # ignore hidden rows and error values
    row_absolute: bool = True,
    column_absolute: bool = True,
    include_sheetname: bool = False,
    external: bool = False,
    formula: bool = False,
) -> str:
    value = _aggregate(
        func,
        ranges,
        option,
        row_absolute=row_absolute,
        column_absolute=column_absolute,
        include_sheetname=include_sheetname,
        external=external,
    )

    if formula:
        return f"={value}"

    return value


# def match_index(ref, sf, columns, column=None, na=False, null=False, error=False):
#     """
#     複数条件にマッチするインデックス(列番号 or 行番号、絶対)を返す数式文字列。

#     Parameters
#     ----------
#     ref : xlviews.SheetFrame
#         検索対象シートフレーム
#     sf : xlviews.SheetFrame
#         検索値を持つシートフレーム
#     columns : str or list of str
#         検索列
#     column : str
#         検索したシートフレームからピックアップするカラムを指定する。
#     na : bool
#         Trueのとき、エラーをNA()で置き換える。
#     null : bool
#         Trueのとき、エラーを""で置き換える。
#     error
#         False以外のとき、errorでエラーを置き換える

#     Returns
#     -------
#     formula : str
#         数式文字列
#     """
#     if isinstance(columns, str):
#         columns = [columns]

#     def gen():
#         sf_columns = sf.columns
#         for column_ in columns:
#             if column_ in sf_columns:
#                 yield sf.range(column_), False
#             else:
#                 yield sf.range(column_, 0)[0], True

#     values, is_wides = zip(*gen(), strict=False)
#     ranges = [ref.range(column_, -1) for column_ in columns]

#     include_sheetname = ranges[0].sheet != values[0].sheet
#     address = "COLUMN" if len(ranges[0].rows) == 1 else "ROW"

#     conditions = []
#     for k, (range_, value, is_wide) in enumerate(
#         zip(ranges, values, is_wides, strict=False),
#     ):
#         range_ = range_.get_address(include_sheetname=include_sheetname)
#         if k == 0:
#             conditions.append(f"{address}({range_})")
#         if is_wide:
#             value = value.get_address(column_absolute=False)
#         else:
#             value = value.get_address(row_absolute=False)
#         condition = "=".join([range_, value])
#         condition = f"({condition})"
#         conditions.append(condition)
#     condition = "*".join(conditions)
#     formula = f"SUMPRODUCT({condition})"

#     if column:
#         cell = ref.range(column, 0)
#         cell = cell.get_address(include_sheetname=include_sheetname)
#         formula = (
#             f"INDIRECT(ADDRESS({formula},COLUMN({cell}),1,1," + f'"{ref.sheet.name}"))'
#         )
#         if error is not False:
#             formula = f"IFERROR({formula},{error})"
#         elif na:
#             formula = f"IFERROR({formula},NA())"
#         elif null:
#             formula = f'IFERROR({formula},"")'
#     return formula


# def interp1d(x, y, value, error='""'):
#     """
#     xの範囲とyの範囲を線形補完する。xは昇順になっていること。

#     Parameters
#     ----------
#     x : xlwings.Range
#         xの範囲
#     y : str or int
#         yのカラム指定
#     value : xlwings.Range
#         新しいx
#     error : str
#         エラー時の値

#     Returns
#     -------
#     formula : str
#         数式文字列
#     """
#     include_sheetname = x.sheet != value.sheet

#     def get_address(range_):
#         return range_.get_address(include_sheetname=include_sheetname)

#     value = value.get_address(row_absolute=False)
#     xstart = get_address(x[0])
#     xend = get_address(x[-1])
#     ystart = f'INDIRECT(ADDRESS(ROW({xstart}),{y},1,1,"{x.sheet.name}"))'
#     x = get_address(x)
#     pre = f"AND({xstart}<={value},{value}<={xend})"
#     match = f"MATCH({value},{x})"
#     x = f"OFFSET({xstart},{match}-1,,2)"
#     y = f"OFFSET({ystart},{match}-1,,2)"
#     return f"IF({pre},TREND({y},{x},{value}),{error})"


# def linear_fit(sf, x, y, to=None, a="a", b="b", by=None):
#     """
#     線形フィッティングを求める。

#     Parameters
#     ----------
#     sf : xlviews.SheetFrame
#     x, y: str
#         カラム名
#     to : xlviews.SheetFrame
#         結果を記入するSheetFrame
#     a, b: str
#         カラム名
#     by : str or list of str
#         グルーピング
#     """
#     grouped = sf.groupby(by)
#     xindex = sf.index(x)
#     yindex = sf.index(y)
#     if to is not None:
#         if a not in to.columns:
#             to[a] = 0
#         a = to.range(a)
#         if b not in to.columns:
#             to[b] = 0
#         b = to.range(b)

#     for k, value in enumerate(grouped.values()):
#         if len(value) != 1:
#             raise ValueError("連続範囲のみ可能")
#         x = sf.sheet.range((value[0][0], xindex), (value[0][1], xindex))
#         y = sf.sheet.range((value[0][0], yindex), (value[0][1], yindex))
#         if len(x) > 1:
#             x = x.get_address()
#             y = y.get_address()
#             formula = f"IFERROR(SLOPE({y},{x}),NA())"
#             a.offset(k).value = "=" + formula
#             formula = f"IFERROR(INTERCEPT({y},{x}),NA())"
#             b.offset(k).value = "=" + formula
#         else:
#             y = y.get_address()
#             a.offset(k).value = "0"
#             b.offset(k).value = "=" + y


# def main():
#     import xlviews as xv

#     sf = xv.SheetFrame(2, 2, style=False, index.nlevels=1)
#     to = xv.SheetFrame(2, 6, style=False, index.nlevels=0)
#     linear_fit(sf, "x", "y", to, by="k")
#     # columns = ['time', 'soa%']
#     # y = match_index(sf, ref, columns=columns)
#     # x = ref.column_range(0, -1)
#     # value = sf.column_range('rate')
#     # formula = interp1d(x, y, value)
#     # sf['delta'] = '=' + formula


# if __name__ == "__main__":
#     main()



import pytest

from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize("index", range(1, 1000, 50))
def test_column_name(index: int):
    from xlviews.core.address import column_name_to_index, index_to_column_name

    assert column_name_to_index(index_to_column_name(index)) == index



"""Set styles for Range."""

from __future__ import annotations

from xlwings.constants import LineStyle, MarkerStyle, ScaleType

from xlviews.colors import rgb
from xlviews.config import rcParams
from xlviews.style import set_font_api

COLORS = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]

MARKER_DICT: dict[str, int] = {
    "": MarkerStyle.xlMarkerStyleNone,
    "o": MarkerStyle.xlMarkerStyleCircle,
    "^": MarkerStyle.xlMarkerStyleTriangle,
    "s": MarkerStyle.xlMarkerStyleSquare,
    "d": MarkerStyle.xlMarkerStyleDiamond,
    "+": MarkerStyle.xlMarkerStylePlus,
    "x": MarkerStyle.xlMarkerStyleX,
    ".": MarkerStyle.xlMarkerStyleDot,
    "-": MarkerStyle.xlMarkerStyleDash,
    "*": MarkerStyle.xlMarkerStyleStar,
}

LINE_DICT: dict[str, int] = {
    "": LineStyle.xlLineStyleNone,
    "-": LineStyle.xlContinuous,
    "--": LineStyle.xlDash,
    "-.": LineStyle.xlDashDot,
    ".": LineStyle.xlDot,
}


def get_marker_style(marker: int | str | None) -> int | None:
    if isinstance(marker, int | None):
        return marker

    return MARKER_DICT[marker]


def get_line_style(line: int | str | None) -> int | None:
    if isinstance(line, int | None):
        return line

    return LINE_DICT[line]


def get_axis_label(axis) -> str | None:  # noqa: ANN001
    if not axis.HasTitle:
        return None

    return axis.AxisTitle.Text


def set_axis_label(
    axis,  # noqa: ANN001
    label: str | None = None,
    name: str | None = None,
    size: float | None = None,
    **kwargs,
) -> None:
    if not label:
        axis.HasTitle = False
        return

    axis.HasTitle = True
    axis_title = axis.AxisTitle
    axis_title.Text = label

    name = name or rcParams["chart.font.name"]
    size = size or rcParams["chart.axis.title.font.size"]

    set_font_api(axis_title, name, size=size, **kwargs)


def get_ticks(axis) -> tuple[float, float, float, float]:  # noqa: ANN001
    return (
        axis.MinimumScale,
        axis.MaximumScale,
        axis.MajorUnit,
        axis.MinorUnit,
    )


def set_ticks(
    axis,  # noqa: ANN001
    *args,
    min: float | None = None,  # noqa: A002
    max: float | None = None,  # noqa: A002
    major: float | None = None,
    minor: float | None = None,
    gridlines: bool = True,
) -> None:
    args = [*args, None, None, None, None][:4]

    min = min or args[0]  # noqa: A001
    max = max or args[1]  # noqa: A001
    major = major or args[2]
    minor = minor or args[3]

    if min is not None:
        axis.MinimumScale = min

    if max is not None:
        axis.MaximumScale = max

    if major is not None:
        axis.MajorUnit = major

        if gridlines:
            axis.HasMajorGridlines = True
        else:
            axis.HasMajorGridlines = False

    if minor is not None:
        axis.MinorUnit = minor

        if gridlines:
            axis.HasMinorGridlines = True
        else:
            axis.HasMinorGridlines = False

    if min:
        axis.CrossesAt = min


def set_tick_labels(
    axis,  # noqa: ANN001
    name: str | None = None,
    size: float | None = None,
    number_format: str | None = None,
) -> None:
    name = name or rcParams["chart.font.name"]
    size = size or rcParams["chart.axis.ticklabels.font.size"]
    set_font_api(axis.TickLabels, name, size=size)

    if number_format:
        axis.TickLabels.NumberFormatLocal = number_format


def get_axis_scale(axis) -> str:  # noqa: ANN001
    if axis.ScaleType == ScaleType.xlScaleLogarithmic:
        return "log"

    if axis.ScaleType == ScaleType.xlScaleLinear:
        return "linear"

    raise NotImplementedError


def set_axis_scale(axis, scale: str) -> None:  # noqa: ANN001
    if scale == "log":
        axis.ScaleType = ScaleType.xlScaleLogarithmic
        return

    if scale == "linear":
        axis.ScaleType = ScaleType.xlScaleLinear
        return

    raise NotImplementedError


def set_dimensions(
    api,  # noqa: ANN001
    left: float | None = None,
    top: float | None = None,
    width: float | None = None,
    height: float | None = None,
) -> None:
    if left is not None:
        api.Left = left

    if top is not None:
        api.Top = top

    if width is not None:
        api.Width = width

    if height is not None:
        api.Height = height


def set_area_format(
    api,  # noqa: ANN001
    border: str | int | tuple[int, int, int] | None = None,
    fill: str | int | tuple[int, int, int] | None = None,
    alpha: float | None = None,
) -> None:
    if border is not None:
        api.Format.Line.Visible = True
        api.Format.Line.ForeColor.RGB = rgb(border)

    if fill is not None:
        api.Format.Fill.Visible = True
        api.Format.Fill.ForeColor.RGB = rgb(fill)

    if alpha is not None:
        api.Format.Line.Transparency = alpha
        api.Format.Fill.Transparency = alpha



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture
def ax(sheet: Sheet):
    ct = ChartType.xlXYScatterLines
    return Axes(2, 2, ct, sheet)


def test_add_series_xy(ax: Axes):
    x = ax.sheet.range("A1:A10")
    y = ax.sheet.range("B1:B10")

    x.options(transpose=True).value = list(range(10))
    y.options(transpose=True).value = list(range(10, 20))
    s = ax.add_series(x, y)

    assert s.api.XValues == (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    assert s.x == (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    assert s.api.Values == (10, 11, 12, 13, 14, 15, 16, 17, 18, 19)
    assert s.y == (10, 11, 12, 13, 14, 15, 16, 17, 18, 19)

    x.options(transpose=True).value = list(range(20, 30))
    y.options(transpose=True).value = list(range(30, 40))

    assert s.api.XValues == (20, 21, 22, 23, 24, 25, 26, 27, 28, 29)
    assert s.x == (20, 21, 22, 23, 24, 25, 26, 27, 28, 29)
    assert s.api.Values == (30, 31, 32, 33, 34, 35, 36, 37, 38, 39)
    assert s.y == (30, 31, 32, 33, 34, 35, 36, 37, 38, 39)
    assert s.api.ChartType == ChartType.xlXYScatterLines
    assert s.chart_type == ChartType.xlXYScatterLines


def test_add_series_x(ax: Axes):
    x = ax.sheet.range("C1:C5")
    x.options(transpose=True).value = list(range(100, 105))
    s = ax.add_series(x)
    assert s.api.XValues == ("1", "2", "3", "4", "5")
    assert s.x == ("1", "2", "3", "4", "5")
    assert s.api.Values == (100, 101, 102, 103, 104)
    assert s.y == (100, 101, 102, 103, 104)


def test_add_series_chart_type(ax: Axes):
    x = ax.sheet.range("D1:D5")
    s = ax.add_series(x, chart_type=ChartType.xlXYScatter)
    assert s.api.ChartType == ChartType.xlXYScatter
    assert s.chart_type == ChartType.xlXYScatter


def test_add_series_name_range(ax: Axes):
    x = ax.sheet.range("A2:A5")
    rng = Range(1, 1, ax.sheet)
    label = rng.get_address(include_sheetname=True, formula=True)
    s = ax.add_series(x, label=label)

    rng.value = "Series Name"
    assert s.api.Name == "Series Name"
    assert s.label == "Series Name"


def test_add_series_xy_range_collection(ax: Axes):
    from xlviews.core.range_collection import RangeCollection

    ax.sheet.range("A1:A10").options(transpose=True).value = list(range(10))
    ax.sheet.range("B1:B10").options(transpose=True).value = list(range(10, 20))

    x = RangeCollection([(1, 3), (8, 10)], 1, ax.sheet)
    y = RangeCollection([(1, 3), (8, 10)], 2, ax.sheet)
    s = ax.add_series(x, y)

    assert s.api.XValues == (0, 1, 2, 7, 8, 9)
    assert s.x == (0, 1, 2, 7, 8, 9)
    assert s.api.Values == (10, 11, 12, 17, 18, 19)
    assert s.y == (10, 11, 12, 17, 18, 19)



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.dist_frame import DistFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.dist_frame import Parent

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    ("columns", "values"),
    [
        (None, ["a_n", "a_v", "a_s", "b_n", "b_v", "b_s"]),
        ("a", ["a_n", "a_v", "a_s"]),
        ("b", ["b_n", "b_v", "b_s"]),
    ],
)
def test_columns(columns, values, sheet: Sheet):
    fc = Parent(sheet, 3, 2)
    sf = DistFrame(fc.sf, columns, by=["x", "y"])
    assert sf.columns.to_list() == values


def test_index_str(sheet: Sheet):
    fc = Parent(sheet, 3, 2)
    sf = DistFrame(fc.sf, by="x")
    assert sf.index.names == ["x"]


def test_index_none(sheet: Sheet):
    fc = Parent(sheet, 3, 2)
    sf = DistFrame(fc.sf)
    assert sf.index.names == [None]
    assert sf.index.to_list() == list(range(14))


def test_group_error(sheet: Sheet):
    df = DataFrame(
        {
            "x": [1, 1, 1, 2, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2],
            "y": [3, 3, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 4, 4],
            "a": [5, 4, 3, 2, 1, 4, 3, 2, 1, 3, 2, 1, 2, 1],
            "b": [1, 2, 3, 4, 5, 1, 2, 3, 4, 1, 2, 3, 1, 2],
        },
    )
    df = df.set_index(["x", "y"])
    sf = SheetFrame(3, 2, data=df, sheet=sheet)
    with pytest.raises(ValueError, match="group must be continuous"):
        sf = DistFrame(sf, None, by=["x", "y"])



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.facet_agg import facet
from xlviews.testing.sheet_frame.pivot import Pivot

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Pivot(sheet_module)


@pytest.fixture(scope="module")
def df(fc: Pivot):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: Pivot):
    return fc.sf


@pytest.fixture(scope="module")
def hfs(sf: SheetFrame):
    return [hf for _, hf in facet(sf)]


@pytest.mark.parametrize("i", [0, 1, 2])
@pytest.mark.parametrize("j", [0, 1])
def test_value(hfs: list[HeatFrame], df: DataFrame, i: int, j: int):
    x = hfs[i].cell.offset(1, j + 4).value
    df = df.reset_index()
    a = df[(df["B"] == i + 1) & (df["Y"] == 1) & (df["A"] == 2) & (df["X"] == j + 2)]
    y = a["u"].mean()
    assert x == y



from __future__ import annotations

from xlviews.core.range import Range
from xlviews.dataframes.colorbar import Colorbar
from xlviews.testing.common import create_sheet

if __name__ == "__main__":
    sheet = create_sheet()

    rng = Range((2, 2), (3, 3))
    rng.value = [[1, 2], [3, 4]]
    cb = Colorbar(2, 5, 6, sheet=sheet)
    cb.set(vmin=rng, vmax=rng, label="T", autofit=True)
    cb.set_adjacent_column_width(1)

    cb = Colorbar(2, 7, 10, orientation="horizontal")
    cb.set(vmin=rng, vmax=rng, label="T", autofit=True)
    cb.set_adjacent_column_width(1)



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, ParamSpec, TypeVar

import xlwings
from pandas import DataFrame, Series
from xlwings.constants import DVType, FormatConditionOperator

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator

    from numpy.typing import NDArray
    from pandas import Index
    from xlwings import Range as RangeImpl

    from xlviews.core.range import Range


def constant(type_: str, name: str | None = None) -> int:
    """Return the Excel constant.

    Args:
        type_ (str): The type name.
        name (str): The name.

    Examples:
        >>> constant("BordersIndex", "EdgeTop")
        8
    """
    if name is None:
        if "." in type_:
            type_, name = type_.split(".")
        else:
            type_, name = "Constants", type_

    if not name.startswith("xl"):
        name = "xl" + name[0].upper() + name[1:]

    type_ = getattr(xlwings.constants, type_)

    return getattr(type_, name)


def iter_columns(
    iterable: Iterable | DataFrame,
    columns: str | list[str],
) -> Iterator[str]:
    """Yield the columns in the order of appearance with colon notation.

    Examples:
        >>> from pandas import DataFrame
        >>> df = DataFrame([[1, 2, 3]], columns=["A", "B", "C"])
        >>> list(iter_columns(df, "B"))
        ['B']
        >>> list(iter_columns(df, ":B"))
        ['A', 'B']
        >>> list(iter_columns(df, "::B"))
        ['A']
    """
    if isinstance(columns, str):
        columns = [columns]

    lst = [x for x in iterable if isinstance(x, str)]

    for c in columns:
        if c.startswith("::"):
            yield from lst[: lst.index(c[2:])]
        elif c.startswith(":"):
            yield from lst[: lst.index(c[1:]) + 1]
        else:
            yield c


def iter_group_locs(
    index: list | NDArray | Index | Series,
    offset: int = 0,
    padding: int = 0,
) -> Iterator[tuple[int, int]]:
    """Yield the group ranges of the index.

    The padding is added to the start and end of each group.
    The end is inclusive.

    Args:
        index: The index to iterate over.
        padding: The padding to add to the start and end of each group.

    Examples:
        >>> list(iter_group_locs([1, 1, 1, 2, 2, 3, 3, 3]))
        [(0, 2), (3, 4), (5, 7)]

        >>> list(iter_group_locs([1, 1, 1, 2, 2, 3, 3, 3], offset=1))
        [(1, 3), (4, 5), (6, 8)]

        >>> list(iter_group_locs([1, 1, 1, 2, 2, 3, 3, 3], padding=3))
        [(0, 2), (6, 7), (11, 13)]
    """
    s = Series(index)
    idx = s[~s.duplicated()].index

    it = zip(idx, [*idx[1:], len(s)], strict=True)
    for k, (start, end) in enumerate(it):
        yield (start + offset + padding * k, end + offset + padding * k - 1)


def add_validate_list(
    rng: Range | RangeImpl,
    value: list[object],
    default: object | None = None,
) -> None:
    if default:
        rng.value = default

    type_ = DVType.xlValidateList
    operator = FormatConditionOperator.xlEqual
    formula = ",".join(map(str, value))

    rng.api.Validation.Add(Type=type_, Operator=operator, Formula1=formula)


P = ParamSpec("P")
R = TypeVar("R")


def suspend_screen_updates(func: Callable[P, R]) -> Callable[P, R]:
    """Suspend screen updates to speed up operations."""

    @wraps(func)
    def _func(*args: P.args, **kwargs: P.kwargs) -> R:
        if app := xlwings.apps.active:
            is_updating = app.screen_updating
            app.screen_updating = False

        try:
            return func(*args, **kwargs)
        finally:
            if app:
                app.screen_updating = is_updating

    return _func


# def label_func_from_list(columns, post=None):
#     """
#     カラム名のリストからラベル関数を作成して返す。

#     Parameters
#     ----------
#     columns : list of str
#         カラム名のリスト
#     post : str, optional
#         追加文字列

#     Returns
#     -------
#     callable
#     """

#     def get_format(t):
#         name_ = f"column.label.{t}"
#         if name_ in rcParams:
#             return rcParams[name_]
#         return "{" + t + "}"

#     fmt_dict = OrderedDict()
#     for column in columns:
#         fmt_dict[column] = get_format(column)

#     def func(**by_key):
#         labels = []
#         for by, fmt in fmt_dict.items():
#             key = by_key[by]
#             if isinstance(fmt, str):
#                 label = fmt.format(**{by: key})
#             else:
#                 label = fmt(key)
#             labels.append(label)
#         return "_".join(labels) + ("_" + post if post else "")

#     return func


# def format_label(data, fmt, sel=None, default=None):
#     dict_ = default.copy() if default else {}
#     if callable(fmt):
#         for column in data.columns:
#             try:
#                 values = data[column]
#             except TypeError:
#                 continue
#             if sel is not None:
#                 values = values[sel]
#             values = values.unique()
#             if len(values) == 1:
#                 dict_[column] = values[0]
#         return fmt(**dict_)
#     keys = re.findall(r"{([\w.]+)(?:}|:)", fmt)
#     for column in keys:
#         if column in data.columns:
#             values = data[column]
#             if sel is not None:
#                 values = values[sel]
#             values = values.unique()
#             if len(values) == 1:
#                 dict_[column] = values[0]
#     for key in keys:
#         if key not in dict_:
#             warnings.warn(
#                 f"タイトル文字列に含まれる'{key}'が、dfに含まれない。",
#             )
#             dict_[key] = "XXX"
#     return fmt.format(**dict_)



from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.figure.plot import Plot
from xlviews.testing.chart import Base
from xlviews.testing.common import create_sheet

if TYPE_CHECKING:
    from collections.abc import Hashable

if __name__ == "__main__":
    sheet = create_sheet()
    fc = Base(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)

    ax = Axes(2, 8)
    data = sf.agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatter)
        .set(label="abc", marker="o", color="blue", alpha=0.6)
    )

    ax = Axes()
    data = sf.groupby("b").agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatterLines)
        .set(label="b={b}", marker=["o", "s"], color={"s": "red", "t": "blue"})
    )

    ax = Axes()
    data = sf.groupby(["b", "c"]).agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatterLines)
        .set(
            label=lambda x: f"{x['b']},{x['c']}",
            marker="b",
            color=("c", ["red", "green"]),
            size=10,
        )
    )

    def m(x: Hashable) -> str:
        if x == "s":
            return "o"
        return "^"

    ax = Axes(left=0)
    data = sf.groupby("b").agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatter)
        .set(label="{b}", marker=m, size=10)
    )

    def c(x: Hashable) -> str:
        if x == ("s", 100):
            return "red"
        return "blue"

    ax = Axes()
    data = sf.groupby(["b", "c"]).agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatter)
        .set(label="{b}_{c}", color=c, marker=("b", m), size=10)
    )



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, TypeVar

import numpy as np
from pandas import DataFrame, Index, MultiIndex, Series
from xlwings import Range as RangeImpl

from xlviews.core.formula import Func, aggregate
from xlviews.core.range import Range
from xlviews.core.range_collection import RangeCollection
from xlviews.utils import iter_columns

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

    from .sheet_frame import SheetFrame

H = TypeVar("H")
T = TypeVar("T")


def to_dict(keys: Iterable[H], values: Iterable[T]) -> dict[H, list[T]]:
    result = {}

    for key, value in zip(keys, values, strict=True):
        result.setdefault(key, []).append(value)

    return result


def create_group_index(
    a: Sequence | Series | DataFrame,
    sort: bool = True,
) -> dict[tuple, list[tuple[int, int]]]:
    if isinstance(a, DataFrame):
        df = a.reset_index(drop=True)
    else:
        df = DataFrame(a).reset_index(drop=True)

    dup = df[df.ne(df.shift()).any(axis=1)]

    start = dup.index.to_numpy()
    end = np.r_[start[1:] - 1, len(df) - 1]

    keys = [tuple(v) for v in dup.to_numpy()]
    values = [(int(s), int(e)) for s, e in zip(start, end, strict=True)]

    index = to_dict(keys, values)

    if not sort:
        return index

    return dict(sorted(index.items()))


def groupby(
    sf: SheetFrame,
    by: str | list[str] | None,
    *,
    sort: bool = True,
) -> dict[tuple, list[tuple[int, int]]]:
    """Group by the specified column and return the group key and row number."""
    if sf.columns.nlevels != 1:
        raise NotImplementedError

    if not by:
        start = sf.row + sf.columns.nlevels
        end = start + len(sf) - 1
        return {(): [(start, end)]}

    if isinstance(by, list) or ":" in by:
        by = list(iter_columns(sf.index.names, by))

    values = sf.index.to_frame()[by]
    index = create_group_index(values, sort=sort)

    offset = sf.row + sf.columns.nlevels
    return {k: [(x + offset, y + offset) for x, y in v] for k, v in index.items()}


class GroupBy:
    sf: SheetFrame
    by: list[str]
    group: dict[tuple, list[tuple[int, int]]]

    def __init__(
        self,
        sf: SheetFrame,
        by: str | list[str] | None = None,
        *,
        sort: bool = True,
    ) -> None:
        self.sf = sf
        self.by = list(iter_columns(sf.index.names, by)) if by else []
        self.group = groupby(sf, self.by, sort=sort)

    def __len__(self) -> int:
        return len(self.group)

    def keys(self) -> Iterator[tuple]:
        yield from self.group.keys()

    def values(self) -> Iterator[list[tuple[int, int]]]:
        yield from self.group.values()

    def items(self) -> Iterator[tuple[tuple, list[tuple[int, int]]]]:
        yield from self.group.items()

    def __iter__(self) -> Iterator[tuple]:
        return self.keys()

    def __getitem__(self, key: tuple) -> list[tuple[int, int]]:
        return self.group[key]

    def index(
        self,
        *,
        as_address: bool = False,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Index:
        if not as_address:
            values = self.keys()
            df = DataFrame(values, columns=self.by)

        else:
            cs = self.sf.index.names
            column = self.sf.column
            idx = [cs.index(c) + column for c in self.by]

            agg = partial(
                self._agg,
                "first",
                row_absolute=row_absolute,
                column_absolute=column_absolute,
                include_sheetname=include_sheetname,
                external=external,
                formula=formula,
            )

            values = {c: agg(i) for c, i in zip(self.by, idx, strict=True)}
            df = DataFrame(values)

        if len(self.by) == 1:
            return Index(df.iloc[:, 0], name=self.by[0])

        return MultiIndex.from_frame(df)

    def agg(
        self,
        func: Func | dict | Sequence[Func] = None,
        columns: str | list[str] | None = None,
        as_address: bool = False,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> DataFrame:
        if self.sf.columns.nlevels != 1:
            raise NotImplementedError

        if isinstance(func, dict):
            columns = list(func.keys())
        elif isinstance(columns, str):
            columns = [columns]

        idx = self.sf.get_indexer(columns)

        if columns is None:
            columns = self.sf.columns.to_list()

        index = self.index(
            as_address=as_address,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )

        agg = partial(
            self._agg,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )

        if isinstance(func, dict):
            it = zip(func.values(), idx, strict=True)
            values = np.array([list(agg(f, i)) for f, i in it]).T
            return DataFrame(values, index=index, columns=columns)

        if func is None or isinstance(func, str | Range | RangeImpl):
            values = np.array([list(agg(func, i)) for i in idx]).T
            return DataFrame(values, index=index, columns=columns)

        values = np.array([list(agg(f, i)) for i in idx for f in func]).T
        columns_ = MultiIndex.from_tuples([(c, f) for c in columns for f in func])
        return DataFrame(values, index=index, columns=columns_)

    def _agg(self, func: Func, column: int, **kwargs) -> Iterator[str]:
        if func == "first":
            func = None
            for row in self.values():
                rng = Range((row[0][0], column), sheet=self.sf.sheet)
                yield aggregate(func, rng, **kwargs)
        else:
            for row in self.values():
                rng = RangeCollection(row, column, self.sf.sheet)
                yield aggregate(func, rng, **kwargs)



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import Index

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Index(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame):
    assert sf.row == 8
    assert sf.column == 2
    assert sf.index.nlevels == 1
    assert sf.columns.nlevels == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 4


def test_columns(sf: SheetFrame):
    assert sf.columns.to_list() == ["a", "b"]


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["name"]


@pytest.mark.parametrize(("x", "b"), [("name", False), ("a", True)])
def test_contains(sf: SheetFrame, x, b):
    assert (x in sf) is b


def test_iter(sf: SheetFrame):
    assert list(sf) == ["a", "b"]


def test_value(sf: SheetFrame, df: DataFrame):
    df_sf = sf.value
    assert df_sf.equals(df.astype(float))
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)


@pytest.mark.parametrize(("column", "index"), [("name", 2), ("a", 3), ("b", 4)])
def test_loc(sf: SheetFrame, column, index):
    assert sf.get_loc(column) == index


@pytest.mark.parametrize(("column", "index"), [(["name", "b"], [2, 4])])
def test_indexer(sf: SheetFrame, column, index):
    assert sf.get_indexer(column) == index


@pytest.mark.parametrize(
    ("column", "offset", "address"),
    [
        ("b", 0, "$D$9"),
        ("a", -1, "$C$8"),
        ("b", -1, "$D$8"),
        ("name", -1, "$B$8"),
        ("a", None, "$C$9:$C$12"),
    ],
)
def test_get_range(sf: SheetFrame, column: str, offset, address):
    assert sf.get_range(column, offset).get_address() == address


def test_get_address(sf: SheetFrame):
    df = sf.get_address(row_absolute=False, column_absolute=False, formula=True)
    assert df.columns.to_list() == ["a", "b"]
    assert df.index.name == "name"
    assert df.index.to_list() == ["x", "x", "y", "y"]
    assert df.to_numpy()[0, 0] == "=C9"


def test_get_adjacent_cell(sf: SheetFrame):
    assert sf.get_adjacent_cell(0).get_address() == "$F$8"



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xlwings
from pandas import DataFrame
from xlwings import Range as RangeImpl

from .address import index_to_column_name

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any, Self

    from pandas._typing import Axes
    from xlwings import Sheet


class Range:
    row: int
    column: int
    row_end: int
    column_end: int
    sheet: Sheet

    def __init__(
        self,
        cell1: tuple[int, int] | int,
        cell2: tuple[int, int] | int | None = None,
        sheet: Sheet | None = None,
    ) -> None:
        self.sheet = sheet or xlwings.sheets.active

        if isinstance(cell1, tuple):
            if not isinstance(cell2, tuple) and cell2 is not None:
                msg = "cell2 must be a tuple or None"
                raise TypeError(msg)

            self.row, self.column = cell1
            self.row_end, self.column_end = cell2 or cell1

        elif isinstance(cell2, int):
            self.row = self.row_end = cell1
            self.column = self.column_end = cell2

        else:
            msg = "cell2 must be an integer"
            raise TypeError(msg)

    @classmethod
    def from_range(cls, rng: RangeImpl) -> Self:
        start = rng.row, rng.column
        end = rng.last_cell.row, rng.last_cell.column
        return cls(start, end, rng.sheet)

    def __len__(self) -> int:
        return (self.row_end - self.row + 1) * (self.column_end - self.column + 1)

    def __iter__(self) -> Iterator[Self]:
        for row in range(self.row, self.row_end + 1):
            for column in range(self.column, self.column_end + 1):
                yield self.__class__((row, column), sheet=self.sheet)

    def __getitem__(self, key: int) -> Self:
        if key < 0:
            key += len(self)

        if key < 0 or key >= len(self):
            raise IndexError("Index out of range")

        row = self.row + key // (self.column_end - self.column + 1)
        column = self.column + key % (self.column_end - self.column + 1)
        return self.__class__((row, column), sheet=self.sheet)

    @property
    def last_cell(self) -> Self:
        return self[-1]

    def __repr__(self) -> str:
        addr = self.get_address(include_sheetname=True, external=True)
        return f"<{self.__class__.__name__} {addr}>"

    def offset(self, row_offset: int = 0, column_offset: int = 0) -> Self:
        return self.__class__(
            (self.row + row_offset, self.column + column_offset),
            (self.row_end + row_offset, self.column_end + column_offset),
            sheet=self.sheet,
        )

    def get_address(
        self,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> str:
        it = iter_addresses(
            self,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )
        return next(it)

    def iter_addresses(
        self,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Iterator[str]:
        return iter_addresses(
            self,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            cellwise=True,
            formula=formula,
        )

    @property
    def impl(self) -> RangeImpl:
        cell1 = (self.row, self.column)
        cell2 = (self.row_end, self.column_end)
        return self.sheet.range(cell1, cell2)

    @property
    def value(self) -> Any:
        return self.impl.value

    @value.setter
    def value(self, value: Any) -> None:
        self.impl.value = value

    @property
    def api(self):  # noqa: ANN201
        return self.impl.api

    @property
    def frame(self) -> FrameRange:
        return FrameRange(
            (self.row, self.column),
            (self.row_end, self.column_end),
            sheet=self.sheet,
        )


def iter_addresses(
    ranges: Range | Iterable[Range],
    *,
    row_absolute: bool = True,
    column_absolute: bool = True,
    include_sheetname: bool = False,
    external: bool = False,
    cellwise: bool = False,
    formula: bool = False,
) -> Iterator[str]:
    if isinstance(ranges, Range):
        ranges = [ranges]

    for rng in ranges:
        for addr in _iter_addresses(
            rng,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            cellwise=cellwise,
        ):
            if formula:
                yield "=" + addr
            else:
                yield addr


def _iter_addresses(
    rng: Range,
    *,
    row_absolute: bool = True,
    column_absolute: bool = True,
    include_sheetname: bool = False,
    external: bool = False,
    cellwise: bool = False,
) -> Iterator[str]:
    rp = "$" if row_absolute else ""
    cp = "$" if column_absolute else ""

    if external:
        prefix = f"[{rng.sheet.book.name}]{rng.sheet.name}!"
    elif include_sheetname:
        prefix = f"{rng.sheet.name}!"
    else:
        prefix = ""

    if cellwise:
        prefix = f"{prefix}{cp}"
        cindex = range(rng.column, rng.column_end + 1)
        cnames = [index_to_column_name(c) for c in cindex]

        rows = [f"{rp}{row}" for row in range(rng.row, rng.row_end + 1)]
        for row in rows:
            for cn in cnames:
                yield f"{prefix}{cn}{row}"

    elif rng.row == rng.row_end and rng.column == rng.column_end:
        yield f"{prefix}{cp}{index_to_column_name(rng.column)}{rp}{rng.row}"

    else:
        start = f"{cp}{index_to_column_name(rng.column)}{rp}{rng.row}"
        end = f"{cp}{index_to_column_name(rng.column_end)}{rp}{rng.row_end}"
        yield f"{prefix}{start}:{end}"


class FrameRange(Range):
    def get_address(
        self,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
        index: Axes | None = None,
        columns: Axes | None = None,
    ) -> DataFrame:
        rp = "$" if row_absolute else ""
        cp = "$" if column_absolute else ""
        f = "=" if formula else ""

        if external:
            prefix = f"{f}[{self.sheet.book.name}]{self.sheet.name}!{cp}"
        elif include_sheetname:
            prefix = f"{f}{self.sheet.name}!{cp}"
        else:
            prefix = f"{f}{cp}"

        cindex = range(self.column, self.column_end + 1)
        cnames = [index_to_column_name(c) for c in cindex]

        rows = [f"{rp}{row}" for row in range(self.row, self.row_end + 1)]
        values = np.array([f"{prefix}{cn}{row}" for row in rows for cn in cnames])
        values = values.reshape(len(rows), len(cnames))

        return DataFrame(values, index=index, columns=columns)



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.core.range import Range
from xlviews.dataframes.heat_frame import Colorbar, HeatFrame
from xlviews.testing.common import create_sheet
from xlviews.testing.sheet_frame.pivot import Pivot

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterator
    from typing import Any, Literal

    from xlviews.dataframes.sheet_frame import SheetFrame


def pair(
    sf: SheetFrame,
    values: str | list[str] | None = None,
    columns: str | list[str] | None = None,
    axis: Literal[0, 1] | None = None,
) -> Iterator[tuple[dict[Hashable, Any], HeatFrame]]:
    sf.set_adjacent_column_width(1)

    rng = sf.get_range("u")
    cbu = Colorbar(3, 11, 12)
    cbu.set(vmin=rng, vmax=rng).autofit()
    rng = sf.get_range("v")
    cbv = Colorbar(16, 11, 12)
    cbv.set(vmin=rng, vmax=rng).autofit()
    cbv.set_adjacent_column_width(1)

    df = sf.pivot_table(["u", "v"], ["B", "Y", "y"], ["A", "X", "x"], formula=True)

    for key, frame in HeatFrame.pair(
        2,
        13,
        df,
        values=values,
        index="B",
        columns=columns,
        axis=axis,
    ):
        frame.autofit()
        frame.set_adjacent_column_width(1)
        if key["value"] == "u":
            cbu.apply(frame.range)
        else:
            cbv.apply(frame.range)
        cell = Range(frame.row - 1, frame.column + 1)
        cell.value = str(key)
        yield key, frame


if __name__ == "__main__":
    sheet = create_sheet()
    fc = Pivot(sheet, style=True)
    # list(pair(fc.sf, values="u", columns=None))
    # list(pair(fc.sf, values=None, columns="A", axis=1))
    list(pair(fc.sf, values="v", columns="A"))



from __future__ import annotations

from typing import TYPE_CHECKING

from .range import Range, iter_addresses

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from xlwings import Sheet


class RangeCollection:
    ranges: list[Range]

    def __init__(
        self,
        row: int | Sequence[int | tuple[int, int]],
        column: int | Sequence[int | tuple[int, int]],
        sheet: Sheet | None = None,
    ) -> None:
        self.ranges = list(_iter_ranges_from_index(row, column, sheet))

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        addr = self.get_address(row_absolute=True, column_absolute=True)
        return f"<{cls} {addr}>"

    def __len__(self) -> int:
        return len(self.ranges)

    def __iter__(self) -> Iterator[Range]:
        return iter(self.ranges)

    def get_address(
        self,
        *,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
    ) -> str:
        it = self.iter_addresses(
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
        )
        return ",".join(it)

    def iter_addresses(
        self,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        cellwise: bool = False,
        formula: bool = False,
    ) -> Iterator[str]:
        return iter_addresses(
            self,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            cellwise=cellwise,
            formula=formula,
        )

    @property
    def api(self):  # noqa: ANN201
        api = self.ranges[0].api

        if len(self.ranges) == 1:
            return api

        sheet = self.ranges[0].sheet
        union = sheet.book.app.api.Union

        for r in self.ranges[1:]:
            api = union(api, r.api)

        return api


def _iter_ranges_from_index(
    row: int | Sequence[int | tuple[int, int]],
    column: int | Sequence[int | tuple[int, int]],
    sheet: Sheet | None = None,
) -> Iterator[Range]:
    if isinstance(row, int) and isinstance(column, int):
        yield Range((row, column), sheet=sheet)

    elif isinstance(row, int) and not isinstance(column, int):
        for c in column:
            start, end = (c, c) if isinstance(c, int) else c
            yield Range((row, start), (row, end), sheet)

    elif isinstance(column, int) and not isinstance(row, int):
        for r in row:
            start, end = (r, r) if isinstance(r, int) else r
            yield Range((start, column), (end, column), sheet)

    else:
        msg = f"Either row or column must be an integer: {row=}, {column=}"
        raise TypeError(msg)



import numpy as np
import pytest
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.base import MultiIndex, MultiIndexParent

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc_parent(sheet_module: Sheet):
    return MultiIndexParent(sheet_module)


@pytest.fixture(scope="module")
def sf(fc_parent: MultiIndexParent):
    fc = MultiIndex(fc_parent.sf)
    return fc.sf


def test_index(sf: HeatFrame):
    x = sf.sheet.range("V3:V26").value
    assert x
    y = np.array([None] * 24)
    y[::6] = range(1, 5)
    np.testing.assert_array_equal(x, y)


def test_index_from_sf(sf: HeatFrame):
    x = np.repeat(list(range(1, 5)), 6)
    np.testing.assert_array_equal(sf.index, x)


def test_columns(sf: HeatFrame):
    x = sf.sheet.range("W2:AH2").value
    assert x
    y = np.array([None] * 12)
    y[::4] = range(1, 4)
    np.testing.assert_array_equal(x, y)


def test_columns_from_sf(sf: HeatFrame):
    x = np.repeat(list(range(1, 4)), 4)
    np.testing.assert_array_equal(sf.columns, x)


@pytest.mark.parametrize(
    ("i", "value"),
    [
        (3, [0, 5, None, 13, 72]),
        (4, [1, None, 9, 14, 73]),
        (5, [None, 6, 10, 15, None]),
    ],
)
def test_values(sf: HeatFrame, i: int, value: int):
    assert sf.sheet.range(f"W{i}:AA{i}").value == value
